import os
import networkx
import itertools
import logging
import collections
import fnmatch
import logging

import pypeliner.helpers
import pypeliner.identifiers
import pypeliner.workflow


class IncompleteJobException(Exception):
    pass


class AmbiguousInputException(Exception):
    def __init__(self, id):
        self.id = id

    def __str__(self):
        return 'temp input {0} not created by any job'.format(self.id)


class AmbiguousOutputException(Exception):
    def __init__(self, output, jobs):
        self.output = output
        self.jobs = jobs

    def __str__(self):
        return 'output {0} created by jobs {1}'.format(self.output, ' and '.join([str(j) for j in self.jobs]))


class DependencyCycleException(Exception):
    def __init__(self, cycle):
        self.cycle = cycle

    def __str__(self):
        return 'dependency cycle {0}'.format(self.cycle)


class NoJobs(Exception):
    pass


class DependencyGraph:
    """ Graph of dependencies between jobs.
    """

    def __init__(self):
        self.completed = set()
        self.created = set()
        self.running = set()
        self.obsolete = set()

    def regenerate(self, jobs):
        """ Create the dependency graph from a set of jobs, and pipeline inputs
        and outputs, maintaining current state.

        """
        self.jobs = jobs
        all_inputs = set((input.id for job in self.jobs.values() for input in job.inputs))
        all_outputs = set((output.id for job in self.jobs.values() for output in job.outputs))
        self.inputs = all_inputs.difference(all_outputs)
        self.outputs = all_outputs.difference(all_inputs)

        self.dependant_jobs = collections.defaultdict(set)
        self.creating_job = dict()
        for job in jobs.values():
            for resource in job.inputs:
                self.dependant_jobs[resource.id].add(job.id)
            for resource in job.outputs:
                if resource.id in self.creating_job:
                    raise AmbiguousOutputException(resource.id, [job.id, self.creating_job[resource.id]])
                self.creating_job[resource.id] = job.id
        for job in jobs.values():
            for resource in job.inputs:
                if resource.id not in self.creating_job and resource.is_temp:
                    raise AmbiguousInputException(resource.id)

        # Create the graph
        G = networkx.DiGraph()
        for job in self.jobs.values():
            # only track job names, ignore the axes value.
            # this will keep DAG smaller and speed up cycle testing
            job_node = ('job', job.jobname)
            if G.has_node(job_node):
                continue
            G.add_node(job_node, job=job)
            chunks = [v.chunk for v in job.node]
            for input in job.inputs:
                input_chunks = [v.chunk for v in input.node]
                if not input_chunks == chunks:
                    continue
                resource_node = ('resource', input.id)
                if resource_node not in G:
                    G.add_node(resource_node, resource=input)
                G.add_edge(resource_node, job_node)
            for output in job.outputs:
                output_chunks = [v.chunk for v in output.node]
                if not output_chunks == chunks:
                    continue
                resource_node = ('resource', output.id)
                if resource_node not in G:
                    G.add_node(resource_node, resource=output)
                G.add_edge(job_node, resource_node)

        # Check for cycles
        try:
            cycles = networkx.find_cycle(G)
        except networkx.NetworkXNoCycle:
            cycles = []
        if len(cycles) > 0:
            raise DependencyCycleException(cycles[0])

        # Pre-compute traversals of the DAG
        self.jobs_forward = list(self.traverse_jobs_forward())
        self.jobs_reverse = list(self.traverse_jobs_reverse())

        # Assume pipeline inputs exist
        self.created.update(self.inputs)

    def traverse_jobs_forward(self):
        """ Traverse jobs in order of execution.
        """
        created_resources = set(self.inputs)

        adjacent_jobs = set()

        for job in self.jobs.values():
            if len(list(job.inputs)) == 0:
                adjacent_jobs.add(job.id)

        for resource_id in created_resources:
            for job_id in self.dependant_jobs[resource_id]:
                adjacent_jobs.add(job_id)

        while len(adjacent_jobs) > 0:
            for job_id in list(adjacent_jobs):
                job = self.jobs[job_id]
                if all([i.id in created_resources for i in job.inputs]):
                    yield job
                    adjacent_jobs.remove(job_id)
                    for o in job.outputs:
                        created_resources.add(o.id)
                        for dependent_job_id in self.dependant_jobs[o.id]:
                            adjacent_jobs.add(dependent_job_id)
                    break

    def traverse_jobs_reverse(self):
        """ Traverse jobs in order of execution.
        """
        visited_resources = set(self.outputs)
        visited_jobs = set()

        adjacent_resources = set()
        adjacent_jobs = set()

        for job in self.jobs.values():
            if len(list(job.outputs)) == 0:
                adjacent_jobs.add(job.id)

        for resource_id in visited_resources:
            if resource_id in self.creating_job:
                adjacent_jobs.add(self.creating_job[resource_id])

        while len(adjacent_jobs) > 0 or len(adjacent_resources) > 0:
            for job_id in list(adjacent_jobs):
                job = self.jobs[job_id]
                if all([o.id in visited_resources for o in job.outputs]):
                    yield job
                    adjacent_jobs.remove(job_id)
                    visited_jobs.add(job_id)
                    for i in job.inputs:
                        adjacent_resources.add(i.id)
                    break

            for resource_id in list(adjacent_resources):
                if all([job_id in visited_jobs for job_id in self.dependant_jobs[resource_id]]):
                    adjacent_resources.remove(resource_id)
                    visited_resources.add(resource_id)
                    if resource_id in self.creating_job:
                        adjacent_jobs.add(self.creating_job[resource_id])

    def pop_next_job(self):
        """ Return the id of the next job that is ready for execution.
        """
        for job in self.jobs.values():
            if len(job.inputs) == 0 and job.id not in self.running and job.id not in self.completed:
                self.running.add(job.id)
                return job

        for job_id in self.running:
            if len(self.jobs[job_id].inputs) == 0:
                raise NoJobs()

        resource_out_of_date = set()

        resource_required = set()
        job_required = set()

        for job in self.jobs_forward:
            if job.id in self.completed:
                continue
            inputs_out_of_date = any([i.id in resource_out_of_date for i in job.inputs])
            if job.out_of_date() or inputs_out_of_date:
                for o in job.outputs:
                    resource_out_of_date.add(o.id)
                for i in job.inputs:
                    resource_required.add(i.id)

        for job in self.jobs_reverse:
            if job.id in self.completed:
                continue
            for o in job.outputs:
                if o.id in resource_out_of_date and not o.exists:
                    job_required.add(job.id)
                if o.id in resource_required and not o.exists:
                    job_required.add(job.id)
            if job.id in job_required:
                for i in job.inputs:
                    resource_required.add(i.id)

        for job in self.jobs_forward:
            inputs_created = all([i.id in self.created for i in job.inputs])
            if inputs_created and job.id not in self.running and job.id not in self.completed:
                if job.id in job_required:
                    job.is_required_downstream = True
                self.running.add(job.id)
                return self.jobs[job.id]

        raise NoJobs()

    def notify_completed(self, job_id):
        """ A job was completed, advance current state.
        """
        job = self.jobs[job_id]
        self.running.remove(job.id)
        self.completed.add(job.id)
        for input in job.inputs:
            if all([otherjob_id in self.completed for otherjob_id in self.dependant_jobs[input.id]]):
                self.obsolete.add(input)
        for output in job.outputs:
            if len(self.dependant_jobs[output.id]) == 0:
                self.obsolete.add(output)
        self.created.update([output.id for output in job.outputs])

    @property
    def finished(self):
        return all([output in self.created for output in self.outputs])

    def cleanup_obsolete(self):
        for resource in self.obsolete:
            resource.cleanup()
        self.obsolete = set()


class WorkflowInstance(object):
    def __init__(self, workflow_def, db_factory, runskip, node=pypeliner.identifiers.Node(), ctx={}, cleanup=False):
        self._logger = logging.getLogger('pypeliner.workflowgraph')
        self.workflow_def = workflow_def
        self.db_factory = db_factory
        self.runskip = runskip
        self.db = db_factory.create(workflow_def.path_info, node.subdir)
        self.node = node
        self.graph = DependencyGraph()
        self.subworkflows = list()
        self.cleanup = cleanup
        self.regenerate()
        self.ctx = workflow_def.ctx
        if ctx:
            self.ctx.update(ctx)

    def regenerate(self):
        """ Regenerate dependency graph based on job instances.
        """

        jobs = dict()
        for job_inst in self.workflow_def._create_job_instances(self, self.db):
            if job_inst.id in jobs:
                raise ValueError('Duplicate job ' + job_inst.displayname)
            jobs[job_inst.id] = job_inst

        self.graph.regenerate(jobs)

    def finalize_workflows(self):
        """ Finalize any workflows that are finished.
        """

        while len(self.subworkflows) > 0:

            # Pop the next finished workflow if it exists
            try:
                job, workflow = pypeliner.helpers.pop_if(self.subworkflows, lambda j, w: w.finished)
            except IndexError:
                return

            # Complete the workflow
            self.complete_job(job)

    def add_subworkflow(self, job, workflow_def):
        node = self.node + job.node + pypeliner.identifiers.Namespace(job.job_def.name)
        workflow = WorkflowInstance(workflow_def, self.db_factory, self.runskip, node=node, cleanup=self.cleanup)
        self.subworkflows.append((job, workflow))

    def complete_job(self, job):
        self.db.job_shelf[job.displayname] = True
        self.notify_completed(job.id)

    def update_ctx(self, job):
        context_config = pypeliner.helpers.GlobalState.get("context_config")
        runskip = None
        job_displayname = job.displayname
        job_ctx = job.ctx
        if not context_config:
            return job

        contexts = context_config.get('context', {})
        for _,inpctx in contexts.items():
            if fnmatch.fnmatch(job_displayname, inpctx["name_match"]):
                job_ctx.update(inpctx.get("ctx", {}))
                runskip = inpctx.get("runskip")

        job.ctx = job_ctx
        job.runskip_request = runskip
        return job

    def pop_next_job(self):
        """ Pop the next job from the top of this or a subgraph.
        """

        while True:
            # Return any ready jobs from sub workflows
            for job, workflow in self.subworkflows:
                try:
                    return workflow.pop_next_job()
                except NoJobs:
                    pass

            # Finalize finished workflows
            self.finalize_workflows()

            # Remove from self graph if no subgraph jobs
            job = self.graph.pop_next_job()
            job = self.update_ctx(job)

            is_run_required, explaination = self.runskip(job)
            self._logger.info(
                'job ' + job.displayname + ' run: ' + str(is_run_required) + ' explanation: ' + explaination,
                extra={"id": job.displayname, "type": "job", "explanation": explaination, 'task_name': job.id[1]})
            if is_run_required:
                return job
            else:
                self.complete_job(job)
                self._logger.info(
                    'job ' + job.displayname + ' skipped',
                    extra={"id": job.displayname, "type": "job", "status": "skipped", 'task_name': job.id[1]})

    def notify_completed(self, job_id):
        self.graph.notify_completed(job_id)
        if self.cleanup:
            self.graph.cleanup_obsolete()

    @property
    def finished(self):
        return self.graph.finished



import itertools
import os
import time

import pypeliner
import pypeliner.workflow
import pypeliner.managed as mgd


class stuff:
    def __init__(self, some_string):
        self.some_string = some_string

def read_stuff(filename):
    with open(filename, 'r') as f:
        return stuff(''.join(f.readlines()).rstrip())

def split_stuff(stf):
    return dict([(ind,stuff(value)) for ind,value in enumerate(list(stf.some_string))])

def split_file_byline(in_filename, lines_per_file, out_filename_callback, max_files=None):
    with open(in_filename, 'r') as in_file:
        def line_group(line, line_idx=itertools.count()):
            return int(next(line_idx) / lines_per_file)
        for file_idx, lines in itertools.groupby(in_file, key=line_group):
            if max_files is not None and file_idx >= max_files:
                break
            with open(out_filename_callback(file_idx), 'w') as out_file:
                for line in lines:
                    out_file.write(line)

def split_2_file_byline(in_filename, out_filename_1, out_filename_2):
    split_file_byline(in_filename, 2, out_filename_1)
    split_file_byline(in_filename, 2, out_filename_2)

def do_file_stuff(in_filename, out_filename, toadd):
    with open(in_filename, 'r') as in_file, open(out_filename, 'w') as out_file:
        line_number = 0
        for line in in_file:
            out_file.write(str(line_number) + str(toadd) + line.rstrip() + '\n')
            line_number += 1

def merge_file_byline(in_filenames, out_filename):
    with open(out_filename, 'w') as out_file:
        for id, in_filename in sorted(in_filenames.items()):
            with open(in_filename, 'r') as in_file:
                for line in in_file.readlines():
                    out_file.write(line)

def merge_2_file_byline(in_filenames_1, in_filenames_2, out_filename):
    with open(out_filename, 'w') as out_file:
        for in_filenames in (in_filenames_1, in_filenames_2):
            for id, in_filename in sorted(in_filenames.items()):
                with open(in_filename, 'r') as in_file:
                    for line in in_file.readlines():
                        out_file.write(line)

def split_by_line(stf):
    return dict([(ind,stuff(value)) for ind,value in enumerate(stf.some_string.split('\n'))])

def split_by_char(stf):
    return dict([(ind,stuff(value)) for ind,value in enumerate(list(stf.some_string))])

def do_stuff(a):
    return a + '-'

def do_paired_stuff(output_filename, input1_filename, input2_filename):
    os.system('cat ' + input1_filename + ' ' + input2_filename + ' > ' + output_filename)

def dict_arg_stuff(output_filenames, input_filenames):
    append_to_lines(input_filenames['1'], '1', output_filenames['1'])
    append_to_lines(input_filenames['2'], '2', output_filenames['2'])

def merge_stuff(stfs):
    merged = ''
    for split, stf in sorted(stfs.items()):
        merged = merged + stf
    return merged

def write_stuff(a, filename):
    with open(filename, 'w') as f:
        f.write(a)

def append_to_lines(in_filename, append, out_filename):
    with open(in_filename, 'r') as in_file, open(out_filename, 'w') as out_file:
        for line in in_file:
            out_file.write(line.rstrip() + append + '\n')

def append_to_lines_instance(in_filename, instance, out_filename):
    with open(in_filename, 'r') as in_file, open(out_filename, 'w') as out_file:
        for line in in_file:
            out_file.write(line.rstrip() + str(instance) + '\n')

def copy_file(in_filename, out_filename):
    with open(in_filename, 'r') as in_file, open(out_filename, 'w') as out_file:
        for line in in_file:
            out_file.write(line)

def write_list(in_list, out_filename):
    with open(out_filename, 'w') as out_file:
        for a in sorted(in_list):
            out_file.write(str(a))

def do_nothing(*arg):
    pass

def do_assert(*arg):
    assert False

def set_chunks():
    return [1, 2]

def file_transform(in_filename, out_filename, prefix, template_filename, merge_templates):
    with open(template_filename, 'w'):
        pass
    with open(in_filename, 'r') as in_file, open(out_filename, 'w') as out_file:
        for key, value in merge_templates.items():
            out_file.write('{0}\t{1}\n'.format(key, value))
        for line in in_file:
            out_file.write('{0}'.format(prefix) + line)

def write_files(out_filename_callback):
    for chunk in (1, 2):
        with open(out_filename_callback(chunk), 'w') as f:
            f.write('file{0}\n'.format(chunk))

def check_temp(output_filename, temp_filename):
    with open(temp_filename, 'w'):
        pass
    with open(output_filename, 'w') as output_file:
        output_file.write(temp_filename)

def create_workflow_2(input_filename, output_filename):
    workflow = pypeliner.workflow.Workflow()

    workflow.transform(
        name='dofilestuff1',
        func='pypeliner.tests.tasks.do_file_stuff',
        args=(
            mgd.InputFile(input_filename),
            mgd.TempOutputFile('intermediate1'),
            'a'))

    workflow.transform(
        name='dofilestuff2',
        func='pypeliner.tests.tasks.do_file_stuff',
        args=(
            mgd.TempInputFile('intermediate1'),
            mgd.OutputFile(output_filename),
            'b'))

    return workflow

def create_workflow_1(input_filename, output_filename):
    workflow = pypeliner.workflow.Workflow()

    # Read data into a managed object
    workflow.transform(
        name='read',
        func='pypeliner.tests.tasks.read_stuff',
        ret=mgd.TempOutputObj('input_data'),
        args=(mgd.InputFile(input_filename),))

    # Extract a property of the managed object, modify it
    # and store the result in another managed object
    workflow.transform(
        name='do',
        func='pypeliner.tests.tasks.do_stuff',
        ret=mgd.TempOutputObj('output_data'),
        args=(mgd.TempInputObj('input_data').prop('some_string'),))

    # Write the object to an output file
    workflow.transform(
        name='write',
        func='pypeliner.tests.tasks.write_stuff',
        args=(
            mgd.TempInputObj('output_data'),
            mgd.TempOutputFile('output_file')))

    # Recursive workflow
    workflow.subworkflow(
        name='sub_workflow_2',
        func='pypeliner.tests.tasks.create_workflow_2',
        args=(
            mgd.TempInputFile('output_file'),
            mgd.OutputFile(output_filename)))

    return workflow


def touch(f):
    with open(f, 'w'):
        pass
        
        
def checkexists(f):
    assert os.path.exists(f)


def job1(i1, o1, o2, o3):
    checkexists(i1)
    touch(o1)
    touch(o2)
    touch(o3)
    time.sleep(1)


def job2(i1):
    checkexists(i1)
    time.sleep(1)
    return 'data'


def job3(i1, o1):
    checkexists(i1)
    time.sleep(1)
    touch(o1)


def job4(i1, i2, i3, o1):
    checkexists(i1)
    checkexists(i2)
    touch(o1)
    time.sleep(1)


def job5(i1, o1):
    checkexists(i1)
    touch(o1)
    time.sleep(1)


class Test(object):
    fragment_mean = None
    fragment_stddev = None
    
def calculate_fragment_stats(i1):
    checkexists(i1)
    return Test()
    

def sample_gc(o1, i1, i2, i3, i4):
    checkexists(i1)
    touch(o1)


def gc_lowess(i1, o1, o2):
    checkexists(i1)
    touch(o1)
    touch(o2)


def split_table(o1, i1, i2):
    checkexists(i1)
    touch(o1[0])
    touch(o1[1])
    touch(o1[2])


def gc_map_bias(i1, i2, i3, i4, o1, i5, i6):
    checkexists(i1)
    checkexists(i4)
    touch(o1)


def merge_tables(o1, i1):
    touch(o1)


def biased_length(o1, i1):
    checkexists(i1)
    touch(o1)




""" Reattachable flyweight objects

The reattachable flyweight objects are used for persisting state
that transitions through a series of pickling operations.  After unpickling,
state objects are reattached to an owner automatically.  This allows a state
object to be pickled, updated in a remote process, repickled, and then
unpickled in the original process updating the state to the value set
remotely.
"""

import uuid


class UnavailableError(Exception):
    pass


class FlyweightState(object):
    """ Collection of state objects that reattach upon unpickling.
    """
    instances = {}

    def __init__(self, state_container=None):
        if state_container is None:
            state_container = dict()
        self.state_id = str(uuid.uuid4())
        self.instances[self.state_id] = state_container

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        if self.state_id in self.instances:
            del self.instances[self.state_id]

    def __getstate__(self):
        return (self.state_id,)

    def __setstate__(self, state):
        self.state_id, = state

    def create_flyweight(self, key):
        return ReattachableFlyweight(self, key)

    def set(self, key, value):
        if self.state_id not in self.instances:
            raise UnavailableError()
        self.instances[self.state_id][key] = value

    def get(self, key):
        if self.state_id not in self.instances:
            raise UnavailableError()
        return self.instances[self.state_id].get(key)


class ReattachableFlyweight(object):
    """ Reattachable state object.
    """

    def __init__(self, state, key):
        self.state = state
        self.key = key

    def __getstate__(self):
        saved = None
        try:
            saved = getattr(self, 'saved')
        except AttributeError:
            pass
        return (self.state, self.key, saved)

    def __setstate__(self, state):
        self.state, self.key, saved = state
        if saved is not None:
            self.set(saved)

    def set(self, value):
        try:
            self.state.set(self.key, value)
        except UnavailableError:
            self.saved = value

    def get(self):
        try:
            return self.state.get(self.key)
        except UnavailableError:
            pass
        if hasattr(self, 'saved'):
            return self.saved



import logging
import os

import azure.common
import azure.core.exceptions
import azure.storage.blob as azureblob
import azure.storage.blob._shared_access_signature as blob_sas
import datetime
import pypeliner
from azure.common import AzureHttpError
from azure.core.exceptions import ResourceNotFoundError
from azure.identity import ClientSecretCredential
from pypeliner.helpers import Backoff


class UnconfiguredStorageAccountError(Exception):
    pass


class BlobMissing(Exception):
    pass


class UnknownBlobPermission(Exception):
    pass


class WrongStorageClient(Exception):
    pass


class MissingBlobPath(Exception):
    pass


class MissingContainerName(Exception):
    pass


class IncorrectContainerUri(Exception):
    pass


class AzureLoggingFilter(logging.Filter):
    def __init__(self):
        self.filter_keywords = [
            'azure', 'adal-python', 'urllib3', 'msrest', 'msal'
        ]

    def filter(self, rec):
        logname = rec.name.split('.')[0]
        if logname in self.filter_keywords:
            return False
        return True


class BlobStorageClient(object):
    def __init__(
            self, storage_account_name, client_id, tenant_id, secret_key,
            storage_account_token=None,
    ):
        """
        abstraction around all azure blob related methods
        :param storage_account_name:
        :type storage_account_name:
        :param client_id: Azure AD client id (application id)
        :type client_id: str
        :param tenant_id: Azure AD tenant id
        :type tenant_id: str
        :param secret_key: secret key for the app
        :type secret_key: str
        """
        self.account_name = storage_account_name

        self.logger = self.__get_logger(add_azure_filter=True)

        if not storage_account_token:
            storage_account_token = self.__get_storage_account_token(
                client_id, secret_key, tenant_id
            )

        storage_account_url = self.__get_account_url(storage_account_name)

        self.blob_client = azureblob.BlobServiceClient(
            storage_account_url,
            storage_account_token)

    @staticmethod
    def __get_account_url(account_name):
        """
        generate and return oauth account url
        :param account_name: storage account_name
        :type str
        :return: storage account oauth url
        :rtype: str
        """
        oauth_url = "https://{}.blob.core.windows.net".format(
            account_name
        )

        return oauth_url

    def __get_logger(self, add_azure_filter=False):
        """
        generate a logger, optionally add filter to remove all azure sdk logs
        :param add_azure_filter: set the filter if set
        :type add_azure_filter: bool
        :return: python logger object
        :rtype: logging.Logger
        """
        logger = logging.getLogger('pypeliner.execqueue.azure_batch')
        if add_azure_filter:
            for handler in logging.root.handlers:
                handler.addFilter(AzureLoggingFilter())
        return logger

    def __unpack_blob_uri(self, blob_uri):
        blob_uri = self.__format_blob_name(blob_uri)

        blob_uri = blob_uri.split('/')
        storage_account = blob_uri[0]
        container_name = blob_uri[1]
        blob_name = '/'.join(blob_uri[2:])
        blob_name = self.__format_blob_name(blob_name)

        return storage_account, container_name, blob_name

    def __unpack_container_uri(self, container_uri):
        container_uri = self.__format_blob_name(container_uri)

        container_uri_split = container_uri.split('/')

        if not len(container_uri_split) == 2:
            raise IncorrectContainerUri('could not parse container uri: {}'.format(container_uri))

        storage_account = container_uri_split[0]
        container_name = container_uri_split[1]

        return storage_account, container_name

    def __parse_blob_args(self, blob_uri, container_name, blob_name):
        if blob_uri:
            storage_account, container_name, blob_name = self.__unpack_blob_uri(blob_uri)
            if not storage_account == self.account_name:
                err_str = "The storage client intended for {} used for account {}".format(
                    self.account_name, storage_account
                )
                raise WrongStorageClient(err_str)
        else:
            if container_name is None:
                raise MissingContainerName("Container name not specified")
            elif blob_name is None:
                raise MissingBlobPath("Blob Path not specified")

        blob_name = self.__format_blob_name(blob_name)
        return container_name, blob_name

    def __parse_container_args(self, container_uri, container_name):
        if container_uri:
            storage_account, container_name = self.__unpack_container_uri(container_uri)
            if not storage_account == self.account_name:
                err_str = "The storage client intended for {} used for account {}".format(
                    self.account_name, storage_account
                )
                raise WrongStorageClient(err_str)
        else:
            if container_name is None:
                raise MissingContainerName("Container name not specified")

        return container_name

    def __format_blob_name(self, filename):
        """
        blob storage backend expects path that includes
        storage account name, container name and blob name
        without a leading slash. remove leading slash if input
        starts with one.
        :param filename: input filename
        :return: blob backend compatible blob name
        """
        return filename.strip('/')

    def __get_storage_account_token(
            self, client_id, secret_key, tenant_id
    ):
        """
        Uses the azure management package and the active directory
        credentials to fetch the authentication key for a storage
        account from azure key vault. The key must be stored in azure
        keyvault for this to work.
        :param str accountname: storage account name
        """

        token_credential = ClientSecretCredential(tenant_id, client_id, secret_key)

        return token_credential

    @Backoff(exception_type=AzureHttpError, max_backoff=1800, randomize=True)
    def __get_blob_to_path(
            self, container_name, blob_name, destination_file_path,
    ):
        """
        download data from blob storage
        :param container_name: blob container name
        :param blob_name: blob path
        :param destination_file_path: path to download the file to
        :return: azure.storage.blob.baseblobservice.Blob instance with content properties and metadata
        """
        blob_name = self.__format_blob_name(blob_name)
        try:
            blob_client = self.blob_client.get_blob_client(container_name, blob_name)
            with open(destination_file_path, "wb") as my_blob:
                download_stream = blob_client.download_blob()
                download_stream.readinto(my_blob)
            blob = blob_client.get_blob_properties()
        except Exception as exc:
            self.logger.exception("Error downloading {} from {}".format(blob_name, container_name))
            raise exc

        if not blob:
            self.logger.exception('Blob SDK returned None, retrying')
            raise AzureHttpError('Blob download failure', 1)

        return blob

    def __get_blob_permission_obj(self, permissions):
        permissions = permissions.split('|')
        permissions = {val: True for val in permissions}
        return azure.storage.blob.BlobSasPermissions(**permissions)

    def __get_blob_properties(self, container_name, blob_name):
        blob_name = self.__format_blob_name(blob_name)
        try:
            blob_client = self.blob_client.get_blob_client(container_name, blob_name)
            blob = blob_client.get_blob_properties()
        except azure.common.AzureMissingResourceHttpError:
            raise BlobMissing((container_name, blob_name))
        except:
            raise BlobMissing((container_name, blob_name))

        return blob

    def download_to_path(
            self, destination_file_path, container_name=None, blob_name=None, blob_uri=None,
    ):
        """
        download blob to a path on the node
        :param destination_file_path: path to local file
        :type destination_file_path: str
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)

        try:
            blob = self.__get_blob_to_path(
                container_name, blob_name, destination_file_path,
            )
        except azure.common.AzureMissingResourceHttpError:
            raise pypeliner.storage.InputMissingException(blob_name)

        blob_size = self.get_blob_size(container_name, blob_name)
        assert blob_size == blob.size

        file_size = os.path.getsize(destination_file_path)
        if file_size != blob.size:
            raise Exception('file size mismatch for {}:{} compared to {}:{}'.format(
                blob_name, blob.properties.content_length, destination_file_path, file_size))

    def upload_from_file(
            self, file_path, metadata=None, container_name=None, blob_name=None, blob_uri=None
    ):
        """
        :param file_path: path of file to upload
        :type file_path: str
        :param metadata: metadata to attach to blob
        :type metadata: str
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)

        self.create_container(container_name=container_name)

        blob_client = self.blob_client.get_blob_client(container_name, blob_name)

        if os.stat(file_path).st_size > 1024 * 1024 * 1024:
            blob_client.MAX_BLOCK_SIZE = 100 * 1024 * 1024
            with open(file_path, "rb") as stream:
                blob_client.upload_blob(stream, overwrite=True)
        else:
            blob_client.MAX_BLOCK_SIZE = 4 * 1024 * 1024
            with open(file_path, "rb") as stream:
                blob_client.upload_blob(stream, overwrite=True)

        blob_client.set_blob_metadata(metadata=metadata)

    def blob_exists(self, container_name=None, blob_name=None, blob_uri=None):
        """
        check if blob exists
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        :return: True if blob exists
        :rtype: bool
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)

        blob_client = self.blob_client.get_blob_client(container_name, blob_name)

        try:
            blob_client.get_blob_properties()
            return True
        except:
            return False

    def get_blob_size(self, container_name=None, blob_name=None, blob_uri=None):
        """
        get blob size in bytes
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        :return: blob length
        :rtype: int
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        blob = self.__get_blob_properties(container_name, blob_name)
        return blob.size

    def set_metadata(
            self, metadata_key, metadata_value, container_name=None, blob_name=None, blob_uri=None,
    ):
        """
        set a metadata key value pair for a blob
        :param metadata_key: key for metadata pair
        :type metadata_key: str
        :param metadata_value: value for the metadata pair
        :type metadata_value: Any
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        self.blob_client.set_blob_metadata(
            container_name,
            blob_name,
            {metadata_key: metadata_value})

    def get_metadata(self, metadata_key, container_name=None, blob_name=None, blob_uri=None):
        """
        pull value for provided from blob metadata
        :param metadata_key: key to look up
        :type metadata_key: any
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        :return: value of provided key in metadata
        :rtype: any
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        blob = self.__get_blob_properties(container_name, blob_name)
        return blob.metadata.get(metadata_key)

    def get_last_modified(self, container_name=None, blob_name=None, blob_uri=None):
        """
        get time blob was last modified
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        :return: time formatted as string
        :rtype: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        blob = self.__get_blob_properties(container_name, blob_name)
        return blob.last_modified.strftime('%Y/%m/%d-%H:%M:%S')

    def delete_blob(self, container_name=None, blob_name=None, blob_uri=None):
        """
        delete a blob
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        if self.blob_exists(container_name=container_name, blob_name=blob_name):
            self.blob_client.delete_blob(container_name, blob_name)

    def generate_blob_url(
            self, container_name=None, blob_name=None, blob_uri=None,
            add_sas=False, sas_lifetime_hours=120,
            sas_permissions='read'
    ):
        """
        generate url for blob. can also add sas if required
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        :param add_sas: adds SAS to url if set
        :type add_sas: bool
        :param sas_lifetime_hours: SAS lifetime in hours
        :type sas_lifetime_hours: int
        :param sas_permissions: permissions for SAS token
        :type sas_permissions: str
        :return: blob url
        :rtype: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        sas_token = None
        if add_sas:
            sas_token = self.get_blob_sas_token(
                container_name=container_name,
                blob_name=blob_name,
                blob_permissions=sas_permissions,
                lifetime_hours=sas_lifetime_hours
            )

        return self.make_blob_url(container_name, blob_name, sas_token=sas_token)

    def get_blob_sas_token(
            self, container_name=None, blob_name=None, blob_uri=None,
            blob_permissions='read', lifetime_hours=120
    ):
        """
        generate a SAS token
        :param container_name: name of container
        :type container_name: str
        :param blob_name: blob path
        :type blob_name: str
        :param blob_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type blob_uri: str
        :param blob_permissions: permissions for SAS token
        :type blob_permissions: str
        :param lifetime_hours: lifetime of token in hours
        :type lifetime_hours: int
        :return: sas token
        :rtype: str
        """
        container_name, blob_name = self.__parse_blob_args(blob_uri, container_name, blob_name)
        start_time = datetime.datetime.utcnow()
        expiry_time = datetime.datetime.utcnow() + datetime.timedelta(hours=lifetime_hours)

        token = self.blob_client.get_user_delegation_key(
            key_start_time=start_time,
            key_expiry_time=expiry_time
        )

        sas_token = blob_sas.generate_blob_sas(
            self.account_name,
            container_name,
            blob_name,
            user_delegation_key=token,
            permission=self.__get_blob_permission_obj(blob_permissions),
            start=start_time,
            expiry=expiry_time,
        )

        return sas_token

    def get_container_sas_token(self, container_uri=None, container_name=None, blob_permissions='read',
                                lifetime_hours=120):

        """
        Obtains a shared access signature granting the specified permissions to the
        container.
        :param container_name: name of container
        :type container_name: str
        :param container_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type container_uri: str
        :param blob_permissions: permissions for SAS token
        :type blob_permissions: str
        :param lifetime_hours: lifetime of token in hours
        :type lifetime_hours: int
        :return: A SAS token granting the specified permissions to the container.
        :rtype: str
        """
        container_name = self.__parse_container_args(container_uri, container_name)

        start_time = datetime.datetime.utcnow()
        expiry_time = datetime.datetime.utcnow() + datetime.timedelta(hours=lifetime_hours)

        token = self.blob_client.get_user_delegation_key(
            key_start_time=start_time,
            key_expiry_time=expiry_time
        )

        sas_token = blob_sas.generate_container_sas(
            self.account_name,
            container_name,
            user_delegation_key=token,
            permission=self.__get_blob_permission_obj(blob_permissions),
            start=start_time,
            expiry=expiry_time,
        )

        return sas_token

    def list_blobs(self, container_uri=None, container_name=None, prefix=None):
        """
        list all blobs in a container
        :param container_name: name of container
        :type container_name: str
        :param container_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type container_uri: str
        :param prefix: list blobs that start with this prefix only
        :type prefix: str
        :return: generator to list the blobs
        :rtype: generator
        """
        container_name = self.__parse_container_args(container_uri, container_name)
        blob_client = self.blob_client.get_container_client(container_name)
        container_blobs = blob_client.list_blobs(name_starts_with=prefix)

        for container in container_blobs:
            yield container

    def create_container(self, container_uri=None, container_name=None):
        """
        create a container if it doesnt exist
        :param container_name: name of container
        :type container_name: str
        :param container_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type container_uri: str
        """
        container_name = self.__parse_container_args(container_uri, container_name)

        if not self.container_exists(container_name=container_name):
            container_client = self.blob_client.get_container_client(container_name)
            container_client.create_container()

    def container_exists(self, container_uri=None, container_name=None):
        """
        check if a container exists
        :param container_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type container_uri: str
        :param container_name: name of container
        :type container_name: str
        """
        container_name = self.__parse_container_args(container_uri, container_name)
        container_client = self.blob_client.get_container_client(container_name)
        try:
            container_client.get_container_properties()
            return True
        except ResourceNotFoundError:
            raise

    def make_blob_url(self, container_name, blob_name, sas_token=None):
        protocol = "https"
        primary_endpoint = "{}.blob.core.windows.net".format(self.account_name)

        url = '{}://{}/{}/{}'.format(
            protocol,
            primary_endpoint,
            container_name,
            blob_name,
        )

        if sas_token:
            url += '?' + sas_token
        return url

    def generate_container_url(
            self, container_uri=None, container_name=None, add_sas=True,
            sas_lifetime_hours=120, sas_permissions='read'
    ):
        """
        generate url for the container, adds SAS if required
        :param container_uri: unique blob identifier. generated by joining
        storage account name, container name and blob path
        :type container_uri: str
        :param container_name: name of container
        :type container_name: str
        :param add_sas: adds SAS to url if set
        :type add_sas: bool
        :param sas_lifetime_hours: SAS lifetime in hours
        :type sas_lifetime_hours: int
        :param sas_permissions: permissions for SAS token
        :type sas_permissions: str
        :return: blob url
        :rtype: str
        """
        container_name = self.__parse_container_args(container_uri, container_name)
        sas_token = None
        if add_sas:
            sas_token = self.get_container_sas_token(
                container_name=container_name,
                blob_permissions=sas_permissions,
                lifetime_hours=sas_lifetime_hours
            )

        # WORKAROUND: for some reason there is no function to create an url for a
        # container
        output_sas_url = self.make_blob_url(
            container_name, 'null',
            sas_token=sas_token).replace('/null', '')

        return output_sas_url



"""
Pipeline application functionality

Provides classes and functions to help creation of a pipeline application.
Using this module, it should be possible to create a pipeline application
with the full functionality of pypeliner with only a few lines of code.
Typically, the first few lines of a pipeline application would be as follows.

Import pypeliner::

    import pypeliner

Create an ``argparse.ArgumentParser`` object to handle command line arguments
including a config, then parse the arguments::

    argparser = argparse.ArgumentParser()
    pypeliner.app.add_arguments(argparser)
    argparser.add_argument('arg1', help='Additional argument 1')
    argparser.add_argument('arg2', help='Additional argument 2')
    args = vars(argparser.parse_args())

Create a :py:class:`pypeliner.app.Pypeline` object passing the arguments to the config parameter::

    pyp = pypeliner.app.Pypeline(config=args)

Create a workflow and run::

    workflow = pypeliner.workflow.Workflow()
    workflow.transform(...)
    workflow.commandline(...)

    pyp.run(workflow)

The following options are supported via the config dictionary argument to
:py:class:`pypeliner.app.Pypeline` or as command line arguments
by calling :py:func:`pypeliner.app.add_arguments` on an
``argparse.ArgumentParser`` object and passing the argument dictionary to
:py:class:`pypeliner.app.Pypeline`.

    tmpdir
        Location of pypeliner temporary files.

    pipelinedir
        Location of pypeliner database, logs.

    loglevel
        Logging level for console messages.  Valid logging levels are:
            - CRITICAL
            - ERROR
            - WARNING
            - INFO
            - DEBUG

    submit
        Type of exec queue to use for job submission.  If available, a default value
        will be obtained from the `$DEFAULT_SUBMITQUEUE` environment variable.

    nativespec
        Required for qsub based job submission, specifies how to request memory from
        the cluster system.  Nativespec is treated as a python style format string
        and uses the job's context to resolve named.  If available, a default value
        will be obtained from the `$DEFAULT_NATIVESPEC` environment variable.

    maxjobs
        The maximum number of jobs to execute in parallel, either on a cluster or
        using subprocess to create multiple processes.

    repopulate
        Recreate all temporary files that may have been cleaned up during a previous
        run in which garbage collection was enabled.  Files may be subsequently
        garbage collected after creation depending on the `nocleanup` option.

    nocleanup
        Do not clean up temporary files.  Does not recreate files that were garbage
        collected in a previous run of the pipeline.

    interactive
        Run the pipeline in interactive mode, prompting at each step as to whether the job
        should be rerun.

    sentinel_only
        Run the pipeline in sentinal only mode, no time stamp checking on files for out
        of date status of jobs, rerun jobs based on whether they have already been
        run.

    context_config
        Run jobs within a specific type of container, either docker or singularity.
        config contains credentials for docker or path to dir with singularity containers

"""

import argparse
import datetime
import logging
import os
from collections import *

import pypeliner.execqueue
import pypeliner.execqueue.factory
import pypeliner.helpers
import pypeliner.runskip
import pypeliner.scheduler
import pypeliner.storage
import yaml

ConfigInfo = namedtuple('ConfigInfo', ['name', 'type', 'default', 'help'])

log_levels = ('CRITICAL', 'ERROR', 'WARNING', 'INFO', 'DEBUG')

default_submit_queue = os.environ.get('DEFAULT_SUBMITQUEUE', None)
default_submit_config = os.environ.get('DEFAULT_SUBMITCONFIG', None)
default_nativespec = os.environ.get('DEFAULT_NATIVESPEC', '')
default_storage_type = os.environ.get('DEFAULT_STORAGETYPE', 'local')
default_storage_config = os.environ.get('DEFAULT_STORAGECONFIG', None)

config_infos = list()
config_infos.append(ConfigInfo('tmpdir', str, './tmp', 'location of temporary files'))
config_infos.append(ConfigInfo('pipelinedir', str, None, 'location of pipeline files'))
config_infos.append(ConfigInfo('loglevel', log_levels, log_levels[2], 'logging level for console messages'))
config_infos.append(ConfigInfo('submit', str, default_submit_queue, 'job submission system'))
config_infos.append(ConfigInfo('submit_config', str, default_submit_config, 'job submission system config file'))
config_infos.append(ConfigInfo('nativespec', str, default_nativespec, 'qsub native specification'))
config_infos.append(ConfigInfo('storage', str, default_storage_type, 'file storage system'))
config_infos.append(ConfigInfo('storage_config', str, default_storage_config, 'file storage system config file'))
config_infos.append(ConfigInfo('maxjobs', int, 1, 'maximum number of parallel jobs'))
config_infos.append(ConfigInfo('repopulate', bool, False, 'recreate all temporaries'))
config_infos.append(ConfigInfo('rerun', bool, False, 'rerun the pipeline'))
config_infos.append(ConfigInfo('nocleanup', bool, False, 'do not automatically clean up temporaries'))
config_infos.append(ConfigInfo('interactive', bool, False, 'run in interactive mode'))
config_infos.append(ConfigInfo('sentinel_only', bool, False, 'no timestamp checks, sentinal only'))
config_infos.append(ConfigInfo('context_config', str, None, 'container registry credentials and job context overrides'))

config_defaults = dict([(info.name, info.default) for info in config_infos])


def _add_argument(argparser, config_info):
    args = list()
    kwargs = dict()
    args.append('--' + config_info.name)
    kwargs['default'] = argparse.SUPPRESS
    if type(config_info.type) == tuple:
        kwargs['choices'] = config_info.type
    elif config_info.type == bool:
        kwargs['action'] = 'store_true'
        kwargs['default'] = False
    else:
        kwargs['type'] = config_info.type
    kwargs['help'] = config_info.help
    argparser.add_argument(*args, **kwargs)


def add_arguments(argparser):
    """ Add pypeliner arguments to an argparser object

    :param argparser: ``argparse.ArgumentParser`` object to which pypeliner specific
                      arguments should be added.  See
                      :py:mod:`pypeliner.app` for available options.

    Add arguments to an ``argparse.ArgumentParser`` object to allow command line control
    of pypeliner.  The options should be extracted after calling
    ``argparse.ArgumentParser.parse_args``, converted to a dictionary using `vars`
    and provided to the initializer of a :py:mod:`pypeliner.app.Pypeline`
    object.

    """
    group = argparser.add_argument_group('pypeliner arguments')
    for config_info in config_infos:
        _add_argument(group, config_info)


def load_config(configfile):
    if not configfile:
        return
    with open(configfile) as configyaml:
        return yaml.safe_load(configyaml)


class Pypeline(object):
    """ Pipeline application helper class

    :param modules: list of modules pypeliner must import before running functions
                    for this pipeline.  These modules will be imported prior to
                    calling any of the user functions added to the pipeline.
    :param config: dictionary of configuration options.  See
                   :py:mod:`pypeliner.app` for available options.

    The Pipeline class sets up logging, creates an execution queue, and creates
    the scheduler with options provided by the config argument.

    """

    def __init__(self, modules=(), config=None):
        self.modules = modules
        self.config = dict([(info.name, info.default) for info in config_infos])
        if config is not None:
            self.config.update(config)

        # If no pipelinedir is specified, fall back to previous
        # functionality of having pipeline files in tmpdir, and
        # temporary files in a tmp subdirectory of tmpdir
        if self.config['pipelinedir'] is None:
            self.config['pipelinedir'] = self.config['tmpdir']
            self.config['tmpdir'] = os.path.join(self.config['tmpdir'], 'tmp')

        datetime_log_prefix = datetime.datetime.now().strftime('%Y%m%d-%H%M%S')
        self.logs_dir = os.path.join(self.config['pipelinedir'], 'log', datetime_log_prefix)
        self.pipeline_log_filename = os.path.join(self.logs_dir, 'pipeline.log')
        pypeliner.helpers.makedirs(self.logs_dir)
        pypeliner.helpers.symlink(self.logs_dir, os.path.join(self.config['pipelinedir'], 'log', 'latest'))

        logging.basicConfig(level=logging.DEBUG, filename=self.pipeline_log_filename, filemode='a')
        console = logging.StreamHandler()
        console.setLevel(self.config['loglevel'])
        logging.getLogger('').addHandler(console)
        logfmt = pypeliner.helpers.MultiLineFormatter('%(asctime)s - %(name)s - %(levelname)s - ')
        for handler in logging.root.handlers:
            handler.setFormatter(logfmt)

        # add json log file to the log_dir
        json_log_file = os.path.join(self.logs_dir, 'pipeline.json')
        jsonlog = logging.FileHandler(json_log_file, mode='a')
        jsonlog.setLevel(self.config['loglevel'])
        jsonlog.addFilter(logging.Filter('pypeliner'))
        logfmt = pypeliner.helpers.JsonFormatter()
        jsonlog.setFormatter(logfmt)
        logging.getLogger('').addHandler(jsonlog)

        self.exec_queue = pypeliner.execqueue.factory.create(
            self.config['submit'], modules=self.modules,
            native_spec=self.config['nativespec'],
            config_filename=self.config['submit_config'])

        self.file_storage = pypeliner.storage.create(
            self.config['storage'], workflow_dir=self.config['pipelinedir'])

        self.sch = pypeliner.scheduler.Scheduler()

        self.sch.temps_dir = self.config['tmpdir']
        self.sch.workflow_dir = self.config['pipelinedir']
        self.sch.logs_dir = self.logs_dir
        self.sch.max_jobs = int(self.config['maxjobs'])
        self.sch.cleanup = not self.config['nocleanup']
        pypeliner.helpers.GlobalState.set('context_config', load_config(self.config['context_config']))
        pypeliner.helpers.GlobalState.set('sentinel_only', self.config['sentinel_only'])

        if self.config['sentinel_only']:
            self.runskip = pypeliner.runskip.SentinalRunSkip(
                rerun=self.config['rerun'])
        else:
            self.runskip = pypeliner.runskip.BasicRunSkip(
                repopulate=self.config['repopulate'],
                rerun=self.config['rerun'])

        if self.config['interactive']:
            self.runskip = pypeliner.runskip.InteractiveRunSkip(self.runskip)

    def run(self, workflow):
        with self.exec_queue, self.file_storage:
            try:
                self.sch.run(workflow, self.exec_queue, self.file_storage, self.runskip)
            finally:
                self.runskip.close()
                print ('log file: ' + self.pipeline_log_filename)



import collections
import os

import six

AxisInstanceBase = collections.namedtuple('AxisInstanceBase', ['axis', 'chunk'])


class AxisInstance(AxisInstanceBase):
    def __new__(cls, axis, chunk):
        assert isinstance(axis, six.string_types)
        return super(AxisInstance, cls).__new__(cls, axis, chunk)

    @property
    def subdir(self):
        return os.path.join(self.axis, str(self.chunk))

    @property
    def displayname(self):
        return ':'.join([self.axis, str((self.chunk, '?')[self.chunk is None])])

    @property
    def undefined(self):
        return self.chunk is None


class Namespace(str):
    @property
    def subdir(self):
        return self

    @property
    def displayname(self):
        return self

    @property
    def undefined(self):
        return False


class Node(tuple):
    def __add__(self, a):
        if isinstance(a, (AxisInstance, Namespace)):
            return Node(self + Node([a]))
        elif isinstance(a, Node):
            return Node(super(Node, self).__add__(a))
        else:
            raise ValueError('Invalid type ' + str(type(a)) + ' for addition')

    def __getitem__(self, key):
        if isinstance(key, slice):
            return Node(super(Node, self).__getitem__(key))
        return super(Node, self).__getitem__(key)

    def __getslice__(self, i, j):
        return self.__getitem__(slice(i, j))

    @property
    def axes(self):
        return tuple([a[0] for a in self])

    @property
    def subdir(self):
        if len(self) == 0:
            return ''
        return os.path.join(*([a.subdir for a in self]))

    @property
    def displayname(self):
        return '/'.join([a.displayname for a in self])

    @property
    def undefined(self):
        return any([a.undefined for a in self])


def create_undefined_node(axes):
    return Node([AxisInstance(a, None) for a in axes])



from __future__ import absolute_import

import os
import sys
import shutil
import subprocess
import pypeliner
from .objectstorage import AwsSimpleStorageService


def init_storage():
    """
    init S3 handling code
    :return: pypeliner S3 class
    :rtype: pypeliner.contrib.aws.aws_storage.SimpleStorageService
    """
    client = AwsSimpleStorageService()
    return client


def run_delegator(before, after, stdout, stderr, workdir):
    """
    run pypeliner delegator
    :param before: path to job callable pickle
    :type before: str
    :param after: path to output pickle
    :type after: str
    :param stdout: file with stdout from pypeliner_delegate
    :type stdout: str
    :param stderr: file with stderr from pypeliner_delegate
    :type stderr: str
    :param workdir: dir where intermediate files will be generated
    :type workdir: str
    """
    before = os.path.abspath(before)
    after = os.path.abspath(after)
    stdout = os.path.abspath(stdout)
    stderr = os.path.abspath(stderr)

    current_path = os.getcwd()

    os.chdir(workdir)

    command = ["pypeliner_delegate", before, after]

    with open(stdout, 'w') as stdout_stream,open(stderr, 'w') as stderr_stream:
        p = subprocess.Popen(command, stdout=stdout_stream, stderr=stderr_stream)
        p.communicate()

    os.chdir(current_path)

def upload_data(client, key, bucket):
    """
    upload file to S3
    :param client: S3 handler
    :type client: pypeliner.contrib.aws.aws_storage.SimpleStorageService
    :param key: object name
    :type key: str
    :param bucket: bucket name
    :type bucket: str
    """

    after_pickle = key + '.after'
    client.upload_from_file(after_pickle, after_pickle, bucket)

    stdout = key + ".stdout"
    client.upload_from_file(stdout, stdout, bucket)

    stderr = key + ".stderr"
    client.upload_from_file(stderr, stderr, bucket)

def cleanup(directories, files):
    """
    remove files
    :param directories: directories to remove
    :type directories: list of str
    :param files: files to remove
    :type files: list of str
    """
    for dirname in directories:
        shutil.rmtree(dirname)

    for filename in files:
        os.remove(filename)


def main():
    """
    download the pickle
    cd into workdir and run delegator
    upload after pickle and logs
    remove all intermediate files
    """
    bucket = sys.argv[1]
    key = sys.argv[2]
    workdir = sys.argv[3]

    pypeliner.helpers.makedirs(workdir)

    client = init_storage()

    if key.startswith('/'):
        key = key[1:]

    # assume we're downloading into current dir
    pypeliner.helpers.makedirs(os.path.dirname(key))

    client.download_to_path(key, bucket, key)

    after_pickle = key + '.after'
    stdout = key + ".stdout"
    stderr = key + ".stderr"
    run_delegator(key, after_pickle, stdout, stderr, workdir)

    upload_data(client, key, bucket)

    cleanup([workdir], [key, after_pickle, stdout, stderr])


if __name__ == '__main__':
    main()


"""
Interface classes used to describe objects managed by the pipeline system.

Objects of each type can be used as placeholder arguments to :py:func:`pypeliner.scheduler.Scheduler.transform` or
:py:func:`pypeliner.scheduler.Scheduler.commandline`.  During pipeline execution, the placeholder argument will be
replaced by the appropriate file or object to create the function or command line arguments.

Managed classes have a common set of parameters:

* name - An identifier for the managed object
* axes - The axes on which the managed object is defined

Axes relate to parallelism.  A managed object with an empty list for axes has a single instance in the system.  A
managed object with a single axis will have as many instances as there are chunks defined for that axis.  Axes can also
be nested to arbitrary depth.

For example, suppose we are running the same analysis on 2 datasets, thus our first axis is 'dataset' with 2 chunks 'A'
and 'B'.  Each dataset is split by line, thus our second axis is 'line'.  Each dataset may have a different number of
lines, and as such the number of chunks for the 'line' axis may be different between dataset 'A' and 'B'.  Thus if
dataset 'A' has 2 lines nad 'B' has 1 line, a managed object defined on the axes 'dataset', 'line' will have the
following instances: ``{'dataset':'A', 'line':1}``, ``{'dataset':'A', 'line':2}``, ``{'dataset':'B', 'line':1}``.

A managed object will resolve to a function or command line argument dependent on whether it is a regular input/output,
merge input or split output.  Regular inputs/outputs have the same axes as the job to which they are given.  Merge
inputs are inputs with a single additional axis, the merge axis.  Split outputs are outputs with a single additional
axis, the split axis.

"""

import pypeliner.arguments
import six


class JobArgMismatchException(Exception):
    def __init__(self, name, axes, node):
        self.name = name
        self.axes = axes
        self.node = node
        self.job_name = 'unknown'

    def __str__(self):
        return 'arg {0} with axes {1} does not match job {2} with axes {3}'.format(
            self.name, self.axes, self.job_name, tuple(a[0] for a in self.node)
        )


class Managed(object):
    """ Interface class used to represent a managed data """

    def __init__(self, name, *axes, **kwargs):
        if name is not None and not isinstance(name, six.string_types):
            raise ValueError('name of argument must be string')
        if type(axes) != tuple:
            raise ValueError('axes must be a tuple')
        self.name = name
        self.axes = axes
        self.kwargs = kwargs

    def create_arg(self, job):

        # add cleanup as arg to override tempspace's cleanup option
        if job.workflow.cleanup is not None and not job.workflow.cleanup:
            self.kwargs["cleanup"] = job.workflow.cleanup

        common = 0
        for node_axis, axis in zip((a[0] for a in job.node), self.axes):
            if node_axis != axis:
                break
            common += 1
        axes_specific = self.axes[common:]
        if len(axes_specific) == 0 and self.normal is not None:
            arg = self.normal(
                job.db, self.name, job.node[:common], direct_write=job.direct_write,
                store_dir=job.store_dir, **self.kwargs
            )
        elif len(axes_specific) > 0 and self.splitmerge is not None:
            arg = self.splitmerge(
                job.db, self.name, job.node[:common], axes_specific,
                direct_write=job.direct_write, store_dir=job.store_dir, **self.kwargs
            )
        else:
            raise JobArgMismatchException(self.name, self.axes, job.node)
        return arg


class Template(Managed):
    """ Represents a name templated by axes 

    `Template` objects will resolve the specified `name` templated by the given
    `axes`.  `name` should be a format string, with named fields that match 
    the names of the axes.  

    For instance, `Template('{case}_details', 'case')` will resolve to the
    strings 'tumour_details' and 'normal_details' if the `case` axis has chunks
    'tumour' and 'normal'.

    :param name: The format string to be resolved by pypeliner.  Each axis
                 should appear at least once as a named field in the format
                 string.
    :param axes: The axes to use to resolve `name`.

    """
    normal = pypeliner.arguments.TemplateArg
    splitmerge = pypeliner.arguments.MergeTemplateArg


class TempSpace(Managed):
    """ Represents a temporary file or directory that is not a dependency

    `TempSpace` objects will resolve to a unique path within pypeliner's
    temporary file space.  If axes are given, a path will be provided for each
    chunk of the given axes.  The file is not checked for existance after the
    job referencing the `TempSpace` has finished execution.

    :param name: The name of the temporary file or directory.  The temp file
                 will have this filename, but different instances for different
                 axis chunks will be located in different directories.
    :param axes: The axes for the file.  This should be identical to the
                 axes of the referencing job.

    """
    normal = pypeliner.arguments.TempSpaceArg
    splitmerge = None


class InputFile(Managed):
    """ Interface class used to represent a user specified managed file input

    `InputFile` objects will resolve the specified `name` templated by the given
    `axes`.  `name` should be a format string, with named fields that match 
    the names of the axes.  The modification time of the file will be used to
    determine if the file has been modified more recently than a job's outputs,
    in order to determine if a job must be run.

    For instance, `InputFile('{case}.bam', 'case')` will resolve to the
    strings 'tumour.bam' and 'normal.bam' if the `case` axis has chunks
    'tumour' and 'normal'.

    :param name: The name of the input file.  Each axis should appear at least
                 once as a named field in the filename.
    :param axes: The axes for the input file.

    For a merge input, `InputFile` will resolve to a dictionary of filenames
    as specified above, with chunks of the merge axis as keys.

    """
    normal = pypeliner.arguments.InputFileArg
    splitmerge = pypeliner.arguments.MergeFileArg


class OutputFile(Managed):
    """ Interface class used to represent a user specified managed file output

    `OutputFile` objects will resolve the specified filename templated by the
    given  axes.  `name` should be a format string, with named fields that match 
    the names of the axes.  An `OutputFile` of the given name and axes is 
    associated with a single job that creates that file.

    For instance, `OutputFile('{case}.bam', 'case')` will resolve to the
    strings 'tumour.bam' and 'normal.bam' if the `case` axis has chunks
    'tumour' and 'normal'.

    :param name: The name of the output file.  Each axis should appear at least
                 once as a named field in the filename.
    :param axes: The axes for the output file.

    For a split output, `OutputFile` will resolve to a callback function taking
    the chunk of the split axis as its only parameter and returning the filename
    for that chunk.

    """
    normal = pypeliner.arguments.OutputFileArg
    splitmerge = pypeliner.arguments.SplitFileArg


class File(Managed):
    """ Interface class used to represent a user specified managed file

    """

    def as_input(self):
        return InputFile(self.name, *self.axes, **self.kwargs)

    def as_output(self):
        return OutputFile(self.name, *self.axes, **self.kwargs)

    def create_arg(self, job):
        raise NotImplementedError('create input or output using as_input or as_output')


class _PropGet(object):
    def __init__(self, prop_name):
        self.prop_name = prop_name

    def __call__(self, a):
        return getattr(a, self.prop_name)


class TempInputObj(Managed):
    """ Interface class used to represent a managed object input

    `TempInputObj` objects will resolve to an object managed by the pipeline
    system.  The contents of the object are used for dependency tracking, as
    described for :py:class:`pypeliner.managed.TempOutputObj`.

    :param name: The name of the object.
    :param axes: The axes for the object.

    For a merge input, `TempInputObj` will resolve to a dictionary of objects
    with chunks of the merge axis as keys.

    """
    normal = pypeliner.arguments.TempInputObjArg
    splitmerge = pypeliner.arguments.TempMergeObjArg

    def prop(self, prop_name):
        """
        Resolve to a property of the object instead of the object itself.

        :param name: The name of the property.
        """
        return TempInputObjExtract(self.name, self.axes, _PropGet(prop_name))

    def extract(self, func):
        """
        Resolve to the return value of the given function called on the object
        rather than the object itself.
        
        :param func: The function to be executed on the object.

        .. admonition:: Warning about state

            The function provided should not have any state as this state 
            cannot be tracked by the dependency system.  Appropriate uses
            are a lambda function that accesses a dictionary entry or 
            performs a fixed calculation.
        """
        return TempInputObjExtract(self.name, self.axes, func)


class TempOutputObj(Managed):
    """ Interface class used to represent a managed object output

    `TempOutputObj` objects are only appropriate as return values for calls 
    to :py:func:`pypeliner.scheduler.Scheduler.transform`.  The object returned
    by the function executed for a transform job will be stored by the pipeline
    using `pickle`.

    If returning a user specified type, it is advisable to add a `__eq__` method.
    Dependency tracking for objects is done by checking if the object has changed
    since the last call that created the object, and will call `__eq__`, or will
    default to comparing `__dict__`.

    :param name: The name of the object.
    :param axes: The axes for the object.

    For a split output, the pipeline system expects a dictionary of objects
    with chunks of the split axis as keys.

    """
    normal = pypeliner.arguments.TempOutputObjArg
    splitmerge = pypeliner.arguments.TempSplitObjArg


class TempInputObjExtract(Managed):
    """ Interface class used to represent a property of a managed
    input object """
    normal = pypeliner.arguments.TempInputObjArg
    splitmerge = pypeliner.arguments.TempMergeObjArg

    def __init__(self, name, axes, func):
        Managed.__init__(self, name, *axes)
        self.kwargs['func'] = func


class TempObj(Managed):
    """ Interface class used to represent a managed object

    """

    def as_input(self):
        return TempInputObj(self.name, *self.axes, **self.kwargs)

    def as_output(self):
        return TempOutputObj(self.name, *self.axes, **self.kwargs)

    def create_arg(self, job):
        raise NotImplementedError('create input or output using as_input or as_output')


class TempInputFile(Managed):
    """ Interface class used to represent a managed temporary file input

    `TempInputFile` objects will resolve to a filename in the temporary file
    space of the pipeline.  Temporary files are subject to garbage collection.

    :param name: The name of the temporary file, basename only (no path
                 information).
    :param axes: The axes for the file.

    For a merge input, `InputFile` will resolve to a dictionary of filenames,
    with chunks of the merge axis as keys.

    """
    normal = pypeliner.arguments.TempInputFileArg
    splitmerge = pypeliner.arguments.TempMergeFileArg


class TempOutputFile(Managed):
    """ Interface class used to represent a managed temporary file output

    `TempOutputFile` objects will resolve to a filename in the temporary file
    space of the pipeline.  Temporary files are subject to garbage collection.

    :param name: The name of the temporary file, basename only (no path
                 information).
    :param axes: The axes for the file.

    For a split output, `TempOutputFile` will resolve to a callback function taking
    the chunk of the split axis as its only parameter and returning the filename
    for that chunk.

    """
    normal = pypeliner.arguments.TempOutputFileArg
    splitmerge = pypeliner.arguments.TempSplitFileArg


class TempFile(Managed):
    """ Interface class used to represent a managed temporary file

    """

    def as_input(self):
        return TempInputFile(self.name, *self.axes, **self.kwargs)

    def as_output(self):
        return TempOutputFile(self.name, *self.axes, **self.kwargs)

    def create_arg(self, job):
        raise NotImplementedError('create input or output using as_input or as_output')


class Instance(Managed):
    """ Interface class used to represent an input chunk associated with the
    job instance

    :param axis: The axis of interest for which to obtain chunk information.

    Resolves to the chunk of its job's instance for a given axis.

    """

    def __init__(self, axis):
        self.axis = axis

    def create_arg(self, job):
        return pypeliner.arguments.InputInstanceArg(job.db, job.node, self.axis)


class InputInstance(Instance):
    pass


class InputChunks(Managed):
    """ Interface class used to represent an input chunk list for a specific axis

    :param axes: The axes of interest for which to obtain a list of chunks.

    `InputChunks` acts similar to a merge.  The specified axes should match
    the axes of its job, with a single additional axis as for a merge.  
    Resolves to a list of chunks for the given 'merge' axis.

    """
    normal = None
    splitmerge = pypeliner.arguments.InputChunksArg

    def __init__(self, *axes):
        Managed.__init__(self, None, *axes)


class OutputChunks(Managed):
    """ Interface class used to represent an output that defines the list of
    chunks for a specific axis

    :param axes: The axes for which chunks will be set.

    `OutputChunks` acts similar to a split object.  `OutputChunks` objects are
    only appropriate as return values for calls to
    :py:func:`pypeliner.scheduler.Scheduler.transform`.  The specified axes should
    match the axes of its job, with a single additional axis as for a split.  The
    pipeline system expects the job function to return a list, which is then
    interpreted as the list of chunks for the given 'split' axis.

    """
    normal = None
    splitmerge = pypeliner.arguments.OutputChunksArg

    def __init__(self, *axes, **kwargs):
        Managed.__init__(self, 'chunks', *axes, **kwargs)


class Chunks(Managed):
    """ Interface class used to represent the list of chunks for a specific axis

    """

    def as_input(self):
        return InputChunks(self.name, *self.axes, **self.kwargs)

    def as_output(self):
        return OutputChunks(self.name, *self.axes, **self.kwargs)

    def create_arg(self, job):
        raise NotImplementedError('create input or output using as_input or as_output')


class OutputWorkflow(Managed):
    """ Interface class used to represent a workflow output

    `TempOutputWorkflow` objects are only appropriate as return values for calls 
    to :py:func:`pypeliner.scheduler.Scheduler.subworkflow`.  The workflow returned
    by the function executed for a subworkflow job will be stored by the pipeline
    using `pickle`.

    :param name: The name of the object.
    :param axes: The axes for the object.

    For a split output, the pipeline system expects a dictionary of workflows
    with chunks of the split axis as keys.

    """
    normal = pypeliner.arguments.OutputWorkflowArg
    splitmerge = None



import subprocess
import sys


class CommandLineException(Exception):
    """ A command produced a non-zero exit code.

    :param args: full set of arguments in failed command line
    :param command: command that failed
    :param returncode: exit code of failed command

    """

    def __init__(self, args, command, returncode):
        self.args = args
        self.command = command
        self.returncode = returncode

    def __str__(self):
        return "Command '%s' with return code %d in command line `%s`" % (
            self.command, self.returncode, " ".join(self.args))


class CommandNotFoundException(Exception):
    """ A command was not found on the path.

    :param args: full set of arguments in failed command line
    :param command: command that could not be found

    """

    def __init__(self, args, command):
        self.args = args
        self.command = command

    def __str__(self):
        return "Command '%s' not found in command line `%s`" % (self.command, " ".join(self.args))


class Callable(object):
    """callable functions and args for pypeliner_delegate
    for running in docker containers

    :param func: function to run in docker
    :param args: arguments

    """

    def __init__(self, func, args, kwargs):
        self.func = func
        self.args = args
        self.kwargs = kwargs

    def __call__(self):
        self.retval = self.func(*self.args, **self.kwargs)


def execute(*args):
    """ Execute a command line

    :param args: executable and command line arguments
    :param kwargs: container keyword arguments

    Execute a command line, and handle pipes between processes and to files.
    The '|', '>' and '<' characters are interpretted as pipes between processes,
    to files and from files in the same way as in bash.  Each process is checked
    for successful completion, with a meaningful exception thrown in the case of
    an error.

    :raises: :py:class:`CommandLineException`, :py:class:`CommandNotFoundException`

    """
    if args.count(">") > 1 or args[0] == ">" or args[-1] == ">":
        raise ValueError("Bad redirect to file")

    if args.count("<") > 1 or args[0] == "<" or args[-1] == "<":
        raise ValueError("Bad redirect from file")

    args = [arg for arg in args if arg != '']
    args = list(str(arg) for arg in args)
    command_list = list(_split_list(args, "|"))

    # Determine input filename for redirect if any
    input = None
    if command_list[0].count("<") == 1:
        input = _get_next(command_list[0], "<")
        command_list[0].remove("<")
        command_list[0].remove(input)

    # Determine output filename for redirect if any
    output = None
    if command_list[-1].count(">") == 1:
        output = _get_next(command_list[-1], ">")
        command_list[-1].remove(">")
        command_list[-1].remove(output)

    # Call process dependent on input / output redirects
    if input and output:
        # Call processes with both input and output
        with open(input, 'rb') as infile, open(output, 'wb') as outfile:
            _call_processes(args, command_list, infile, outfile)
    elif input:
        # Call processes with input
        with open(input, 'rb') as infile:
            _call_processes(args, command_list, infile, sys.stdout)
    elif output:
        # Call processes with output
        with open(output, 'wb') as outfile:
            _call_processes(args, command_list, None, outfile)
    else:
        # Call processes with no redirection
        _call_processes(args, command_list, None, sys.stdout)


def _split_list(items, sep):
    sublist = []
    for elem in items:
        if elem == sep:
            yield list(sublist)
            sublist = []
        else:
            sublist.append(elem)
    yield list(sublist)


def _get_next(items, identifier):
    for elem, nextelem in zip(items, items[1:]):
        if elem == identifier:
            return nextelem


def _call_processes(args, command_list, infile, outfile):
    # List of processes created
    processes = []

    if len(command_list) == 1:

        # Create a single process with no pipeing
        processes.append(_call_process(command_list[0], stdin=infile, stdout=outfile, stderr=sys.stderr))

    else:

        # Create the first process with stdin from infile
        processes.append(_call_process(command_list[0], stdin=infile, stdout=subprocess.PIPE, stderr=sys.stderr))

        # Create intermediate processes with pipes
        for command in command_list[1:-1]:
            processes.append(
                _call_process(command, stdin=processes[-1].stdout, stdout=subprocess.PIPE, stderr=sys.stderr))
            processes[-2].stdout.close()

        # Create final process with stdout as outfile
        processes.append(_call_process(command_list[-1], stdin=processes[-1].stdout, stdout=outfile, stderr=sys.stderr))
        processes[-2].stdout.close()

    # Wait for all process
    for process in processes:
        process.wait()

    # Check return codes
    for idx, process in enumerate(processes):
        if len(args) == 2 and args[0] == "bash":
            args = open(args[1]).readlines()
        else:
            args = args

        if process.returncode != 0:
            raise CommandLineException(args, command_list[idx][0], process.returncode)


def _call_process(command, stdin, stdout, stderr):
    try:
        return subprocess.Popen(command, stdin=stdin, stdout=stdout, stderr=stderr)
    except OSError as e:
        if e.errno == 2:
            raise CommandNotFoundException(command, command[0])
        else:
            raise



import unittest
import shutil
import os
import logging
import time

import pypeliner
import pypeliner.execqueue.factory
import pypeliner.runskip
import pypeliner.scheduler
import pypeliner.workflow
import pypeliner.managed as mgd


script_directory = os.path.dirname(os.path.abspath(__file__))
pipeline_dir = os.path.join(script_directory, 'pipeline')


class scheduler_test(unittest.TestCase):

    input_filename = os.path.join(script_directory, 'scheduler_test.input')
    output_filename = os.path.join(script_directory, 'scheduler_test.output')

    input_n_filename = os.path.join(script_directory, 'scheduler_test.{byfile}.input')
    output_n_filename = os.path.join(script_directory, 'scheduler_test.{byfile}.output')
    output_n_template = os.path.join(script_directory, 'scheduler_test.{byfile}.template')

    input_1_filename = input_n_filename.format(byfile=1)
    input_2_filename = input_n_filename.format(byfile=2)

    output_1_filename = output_n_filename.format(byfile=1)
    output_2_filename = output_n_filename.format(byfile=2)

    log_filename = os.path.join(script_directory, './scheduler_test.log')

    try:
        os.remove(log_filename)
    except OSError:
        pass

    logging.basicConfig(level=logging.DEBUG, filename=log_filename, filemode='a')
    console = logging.StreamHandler()
    console.setLevel(logging.DEBUG)
    console.setFormatter(pypeliner.helpers.MultiLineFormatter('%(asctime)s - %(name)s - %(levelname)s - '))
    console.addFilter(logging.Filter('pypeliner'))
    logging.getLogger('').addHandler(console)

    ctx = {
        'mem':1,
        # 'image': 'amcpherson/pypeliner:latest',
        # 'mounts': ['/Users/amcphers/Projects/pypeliner'],
    }

    def setUp(self):

        try:
            shutil.rmtree(pipeline_dir)
        except:
            pass

        try:
            shutil.rmtree(exc_prefix)
        except:
            pass

        try:
            os.remove(self.output_filename)
        except OSError:
            pass

        for chunk in (1, 2):
            try:
                os.remove(self.output_n_filename.format(**{'byfile':chunk}))
                os.remove(self.output_n_template.format(**{'byfile':chunk}))
            except OSError:
                pass

    def run_workflow(self, workflow, cleanup=None, runskip=None):

        scheduler = pypeliner.scheduler.Scheduler()
        scheduler.workflow_dir = pipeline_dir
        scheduler.temps_dir = os.path.join(pipeline_dir, 'tmp')
        scheduler.max_jobs = 10

        if cleanup is not None:
            scheduler.cleanup = cleanup

        exec_queue = pypeliner.execqueue.factory.create('local', [])
        storage = pypeliner.storage.create('local', pipeline_dir)

        if runskip is None:
            runskip = pypeliner.runskip.BasicRunSkip()

        with exec_queue, storage:
            scheduler.run(workflow, exec_queue, storage, runskip)

    def test_simple_chunks1(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Write a set of output files indexed by axis `byfile`
        workflow.transform(
            name='write_files',
            func='pypeliner.tests.tasks.write_files',
            args=(mgd.OutputFile(self.output_n_filename, 'byfile'),))

        self.run_workflow(workflow)

        for chunk in ('1', '2'):
            with open(self.output_n_filename.format(**{'byfile':chunk}), 'r') as output_file:
                output = output_file.readlines()
                self.assertEqual(output, ['file{0}\n'.format(chunk)])

    def test_simple_chunks2(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Directly set the chunks indexed by axis `byfile`
        workflow.transform(
            name='set_chunks',
            func='pypeliner.tests.tasks.set_chunks',
            ret=mgd.OutputChunks('byfile'))

        workflow.setobj(
            obj=mgd.OutputChunks('byfile', 'axis2'),
            value=['a', 'b'],
            axes=('byfile',))

        # Transform the input files indexed by axis `byfile` to output files
        # also indexed by axis `byfile`
        workflow.transform(
            name='do',
            axes=('byfile',),
            func='pypeliner.tests.tasks.file_transform',
            args=(
                mgd.InputFile(self.input_n_filename, 'byfile'),
                mgd.OutputFile(self.output_n_filename, 'byfile'),
                mgd.InputInstance('byfile'),
                mgd.Template(self.output_n_template, 'byfile'),
                mgd.Template('{byfile}_{axis2}', 'byfile', 'axis2')))

        # Merge output files indexed by axis `byfile` into a single output file
        workflow.transform(
            name='merge',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.InputFile(self.output_n_filename, 'byfile'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow)

        merged_expected = []
        for chunk in ('1', '2'):
            self.assertTrue(os.path.exists(self.output_n_template.format(**{'byfile':chunk})))
            with open(self.output_n_filename.format(**{'byfile':chunk}), 'r') as output_file:
                output = output_file.readlines()
                expected = ['{0}\t{1}_{0}\n'.format(x, chunk) for x in ('a', 'b')]
                expected += [chunk + 'line' + str(line_num) + '\n' for line_num in range(1, 9)]
                self.assertEqual(output, expected)
                merged_expected += expected

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()
            self.assertEqual(output, merged_expected)

    def test_simple_objects(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Read data into a managed object
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename),))

        # Extract a property of the managed object, modify it
        # and store the result in another managed object
        workflow.transform(
            name='do',
            func='pypeliner.tests.tasks.do_stuff',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('input_data').prop('some_string'),))

        # Write the object to an output file
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.write_stuff',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile('output')))

        workflow.set_filenames('output', filename=self.output_filename)

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8-'])

    def test_simple_files(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.do_file_stuff',
            ctx={
                'container_type': 'docker',
                'image': 'amcpherson/pypeliner:latest',
                'mounts': {'pypeliner':'/Users/amcphers/Projects/pypeliner'},
            },
            args=(
                mgd.InputFile(self.input_filename),
                mgd.OutputFile(self.output_filename),
                'a',
            )
        )

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['0aline1\n', '1aline2\n', '2aline3\n', '3aline4\n', '4aline5\n', '5aline6\n', '6aline7\n', '7aline8\n'])

    def test_simple_split_files(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.transform(
            name='split_files',
            func='pypeliner.tests.tasks.split_2_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                mgd.TempOutputFile('output', 'split'),
                mgd.TempOutputFile('output2', 'split', axes_origin=[]),
            ),
        )

        workflow.transform(
            name='merge_files',
            func='pypeliner.tests.tasks.merge_2_file_byline',
            args=(
                mgd.TempInputFile('output', 'split'),
                mgd.TempInputFile('output2', 'split'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n', 'line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n'])

    def test_dict_args(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.transform(
            name='dict_arg_stuff',
            func='pypeliner.tests.tasks.dict_arg_stuff',
            args=(
                {'1':mgd.OutputFile(self.output_1_filename),
                 '2':mgd.OutputFile(self.output_2_filename)},
                {'1':mgd.InputFile(self.input_1_filename),
                 '2':mgd.InputFile(self.input_2_filename)}))

        self.run_workflow(workflow)

        for file_num in (1, 2):
            with open(self.output_n_filename.format(byfile=file_num), 'r') as output_file:
                output = output_file.read()
            expected = 'line1{0}\nline2{0}\nline3{0}\nline4{0}\nline5{0}\nline6{0}\nline7{0}\nline8{0}\n'.format(file_num)
            self.assertEqual(output, expected)

    def test_simple_sub_workflow(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.setobj(obj=mgd.OutputChunks('byfile'), value=(1, 2))

        workflow.transform(
            name='append_to_lines',
            axes=('byfile',),
            func='pypeliner.tests.tasks.append_to_lines',
            args=(
                mgd.InputFile(self.input_n_filename, 'byfile'),
                '#',
                mgd.TempOutputFile('intermediate1', 'byfile')))

        workflow.subworkflow(
            name='sub_workflow_1',
            axes=('byfile',),
            func='pypeliner.tests.tasks.create_workflow_1',
            args=(
                mgd.TempInputFile('intermediate1', 'byfile'),
                mgd.TempOutputFile('intermediate2', 'byfile')))

        workflow.transform(
            name='merge_files',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('intermediate2', 'byfile'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        expected = ['0b0aline1#\n', '1b1aline2#\n', '2b2aline3#\n', '3b3aline4#\n',
            '4b4aline5#\n', '5b5aline6#\n', '6b6aline7#\n', '7b7aline8#-\n',
            '0b0aline1#\n', '1b1aline2#\n', '2b2aline3#\n', '3b3aline4#\n',
            '4b4aline5#\n', '5b5aline6#\n', '6b6aline7#\n', '7b7aline8#-\n']

        self.assertEqual(output, expected)

    def test_specify_input_filename(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # For single axis, only single value is supported
        input_filenames = {
            1:self.input_n_filename.format(byfile=1),
            2:self.input_n_filename.format(byfile=2),
        }

        # Merge a set of input files indexed by axis `byfile`
        workflow.setobj(mgd.OutputChunks('byfile'), (1, 2))
        workflow.transform(
            name='merge_files',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.InputFile('input_files', 'byfile'),
                mgd.OutputFile(self.output_filename)))

        workflow.set_filenames('input_files', 'byfile', fnames=input_filenames)

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n', 'line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n'])

    def test_specify_input_filename_template(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Merge a set of input files indexed by axis `byfile`
        workflow.setobj(mgd.OutputChunks('byfile'), (1, 2))
        workflow.transform(
            name='merge_files',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.InputFile('input_files', 'byfile'),
                mgd.OutputFile(self.output_filename)))

        workflow.set_filenames('input_files', 'byfile', template=self.input_n_filename)

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n', 'line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n'])

    def test_specify_output_filename(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # For single axis, only single value is supported
        output_filenames = {
            1:self.output_n_filename.format(byfile=1),
            2:self.output_n_filename.format(byfile=2),
        }

        # Write a set of output files indexed by axis `byfile`
        workflow.transform(
            name='write_files',
            func='pypeliner.tests.tasks.write_files',
            args=(mgd.OutputFile('output_files', 'byfile'),))

        workflow.set_filenames('output_files', 'byfile', fnames=output_filenames)

        self.run_workflow(workflow)

        for chunk in ('1', '2'):
            with open(self.output_n_filename.format(**{'byfile':chunk}), 'r') as output_file:
                output = output_file.readlines()
                self.assertEqual(output, ['file{0}\n'.format(chunk)])

    def test_specify_output_filename_template(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Write a set of output files indexed by axis `byfile`
        workflow.transform(
            name='write_files',
            func='pypeliner.tests.tasks.write_files',
            args=(mgd.OutputFile('output_files', 'byfile'),))

        workflow.set_filenames('output_files', 'byfile', template=self.output_n_filename)

        self.run_workflow(workflow)

        for chunk in ('1', '2'):
            with open(self.output_n_filename.format(**{'byfile':chunk}), 'r') as output_file:
                output = output_file.readlines()
                self.assertEqual(output, ['file{0}\n'.format(chunk)])

    def test_set_multiple_axes(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Merge a set of input files indexed by axis `byfile`
        workflow.setobj(mgd.OutputChunks('byfile', 'bychar'), ((1, 'a'), (1, 'b'), (2, 'a')))
        workflow.transform(
            name='merge_files',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.InputFile('input_files', 'byfile'),
                mgd.OutputFile(self.output_filename)))

        workflow.set_filenames('input_files', 'byfile', template=self.input_n_filename)

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n', 'line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n'])

    def test_tempfile(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Write the name of the temp file produced by pypeliner
        # into an output file
        workflow.transform(
            name='write_files',
            func='pypeliner.tests.tasks.check_temp',
            args=(
                mgd.OutputFile(self.output_filename),
                mgd.TempSpace('temp_space')))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, [os.path.join(pipeline_dir, 'tmp/temp_space')])

    def test_missing_temporary(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.setobj(mgd.OutputChunks('byfile'), (1, 2))

        workflow.transform(
            name='read',
            axes=('byfile',),
            func='pypeliner.tests.tasks.do_file_stuff',
            args=(
                mgd.InputFile('input_files', 'byfile'),
                mgd.TempOutputFile('temp_files', 'byfile'),
                'extras'))

        workflow.transform(
            name='merge_files',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('temp_files', 'byfile'),
                mgd.OutputFile(self.output_filename)))

        workflow.set_filenames('input_files', 'byfile', template=self.input_n_filename)

        self.run_workflow(workflow)

        time.sleep(1)
        os.utime(self.input_1_filename, None)
        print ('restarting')

        self.run_workflow(workflow)

        time.sleep(1)
        os.utime(self.input_1_filename, None)
        os.remove(self.output_filename)
        print ('restarting again')

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        expected_output = []
        for file_id in range(1, 2+1):
            for line in range(1, 8+1):
                expected_output.append('{}{}line{}\n'.format(line-1, 'extras', line))

        self.assertEqual(output, expected_output)

    def test_missing_temporary2(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.transform(
            name='do_file_stuff_1',
            func='pypeliner.tests.tasks.do_file_stuff',
            args=(
                mgd.InputFile(self.input_filename),
                mgd.TempOutputFile('temp_file1'),
                '1'))

        workflow.transform(
            name='do_file_stuff_2',
            func='pypeliner.tests.tasks.do_file_stuff',
            args=(
                mgd.TempInputFile('temp_file1'),
                mgd.TempOutputFile('temp_file2'),
                '2'))

        workflow.transform(
            name='do_file_stuff_3',
            func='pypeliner.tests.tasks.do_file_stuff',
            args=(
                mgd.TempInputFile('temp_file2'),
                mgd.OutputFile(self.output_filename),
                '3'))

        self.run_workflow(workflow)

        time.sleep(1)
        os.remove(self.output_filename)

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        expected_output = []
        for line in range(1, 8+1):
            expected_output.append('{}3{}2{}1line{}\n'.format(line-1, line-1, line-1, line))

        self.assertEqual(output, expected_output)

    def test_missing_temporary3(self):
        
        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.transform(
            name='bamdisc',
            func='pypeliner.tests.tasks.job1',
            args=(
                mgd.InputFile(self.input_filename),
                mgd.TempOutputFile('stats.txt'),
                mgd.TempOutputFile('reads1.fq.gz'),
                mgd.TempOutputFile('sample1.fq.gz'),
            ),
        )

        workflow.transform(
            name='readstats',
            func='pypeliner.tests.tasks.job2',
            ret=mgd.TempOutputObj('stats'),
            args=(
                mgd.TempInputFile('stats.txt'),
            ),
        )

        workflow.transform(
            name='prepseed_sample',
            func='pypeliner.tests.tasks.job3',
            args=(
                mgd.TempInputFile('sample1.fq.gz'),
                mgd.TempOutputFile('sample.seed'),
            ),
        )

        workflow.transform(
            name='bwtrealign_sample',
            func='pypeliner.tests.tasks.job4',
            args=(
                mgd.TempInputFile('sample.seed'),
                mgd.TempInputFile('sample1.fq.gz'),
                mgd.TempInputObj('stats'),
                mgd.TempOutputFile('samples.align.true'),
            ),
        )

        workflow.transform(
            name='job5',
            func='pypeliner.tests.tasks.do_assert',
            args=(
                mgd.TempInputFile('samples.align.true'),
                mgd.OutputFile(self.output_filename),
            ),
        )

        self.assertRaises(pypeliner.scheduler.PipelineException, self.run_workflow, workflow)

        time.sleep(1)

        print ('restarting')

        for name, job in workflow.job_definitions.items():
            if name != 'job5':
                job.func = 'pypeliner.tests.tasks.do_assert'
            else:
                job.func = 'pypeliner.tests.tasks.job5'

        self.run_workflow(workflow)

    def test_simple_create_all(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Read data into a managed object
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename),))

        # Extract a property of the managed object, modify it
        # and store the result in another managed object
        workflow.transform(
            name='do',
            func='pypeliner.tests.tasks.do_stuff',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('input_data').prop('some_string'),))

        # Write the object to an output file
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.write_stuff',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.TempOutputFile('output_file')))

        self.run_workflow(workflow, cleanup=False)

        with open(os.path.join(pipeline_dir, 'tmp/output_file'), 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8-'])

        self.assertTrue(os.path.exists(os.path.join(pipeline_dir, 'tmp/input_data._o')))
        self.assertTrue(os.path.exists(os.path.join(pipeline_dir, 'tmp/output_data._o')))

    def test_cycle(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Read data into a managed object, but also add a superfluous
        # input argument called `cyclic`, which is generated by a
        # downstream job
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(
                mgd.InputFile(self.input_filename),
                mgd.TempInputObj('cyclic')))

        # Extract a property of the managed object, modify it
        # and store the result in another managed object
        workflow.transform(
            name='do',
            func='pypeliner.tests.tasks.do_stuff',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('input_data').prop('some_string'),))

        # Write the object to an output file, and also include `cyclic`
        # as an output so as to create a cycle in the dependency graph
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.write_stuff',
            ret=mgd.TempOutputObj('cyclic'),
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile(self.output_filename)))

        self.assertRaises(pypeliner.graph.DependencyCycleException, self.run_workflow, workflow)

    def test_commandline_simple(self):

        workflow = pypeliner.workflow.Workflow(default_ctx=self.ctx)

        # Copy input to output file using linux `cat`
        workflow.commandline(
            name='do',
            args=(
                'cat',
                mgd.InputFile(self.input_filename),
                '|',
                'cat',
                '>', mgd.OutputFile(self.output_filename),))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8\n'])

    def test_single_object_split(self):

        workflow = pypeliner.workflow.Workflow(default_ctx=self.ctx)

        # Read data into a managed object, which is a string
        workflow.transform(
            name='read', 
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename),))

        # Split the string into individual characters
        workflow.transform(
            name='splitbychar',
            func='pypeliner.tests.tasks.split_stuff',
            ret=mgd.TempOutputObj('input_data', 'bychar'),
            args=(mgd.TempInputObj('input_data'),))

        # Modify each single character string, appending `-`
        workflow.transform(
            name='do',
            axes=('bychar',),
            func='pypeliner.tests.tasks.do_stuff',
            ret=mgd.TempOutputObj('output_data', 'bychar'),
            args=(mgd.TempInputObj('input_data', 'bychar').prop('some_string'),))

        # Merge the modified strings
        workflow.transform(
            name='mergebychar', 
            func='pypeliner.tests.tasks.merge_stuff',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('output_data', 'bychar'),))

        # Write the modified merged string to an output file
        workflow.transform(
            name='write', 
            func='pypeliner.tests.tasks.write_stuff',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['l-i-n-e-1-\n', '-l-i-n-e-2-\n', '-l-i-n-e-3-\n', '-l-i-n-e-4-\n', '-l-i-n-e-5-\n', '-l-i-n-e-6-\n', '-l-i-n-e-7-\n', '-l-i-n-e-8-'])

    def test_split_getinstance(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Split input file by line and output one file per line
        workflow.transform(
            name='splitbyline',
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                1,
                mgd.TempOutputFile('input_filename', 'byline')))

        # Append the `instance` of the split (basically the index of the line)
        # to each temporary file `input_filename` and output to a new file
        workflow.transform(
            name='append',
            axes=('byline',),
            func='pypeliner.tests.tasks.append_to_lines_instance',
            args=(
                mgd.TempInputFile('input_filename', 'byline'),
                mgd.InputInstance('byline'),
                mgd.TempOutputFile('output_filename', 'byline')))

        # Merge files and output
        workflow.transform(
            name='mergebyline',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('output_filename', 'byline'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line10\n', 'line21\n', 'line32\n', 'line43\n', 'line54\n', 'line65\n', 'line76\n', 'line87\n'])

    def test_split_getinstances(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Split file by line and output a single line per temporary output
        # file named `input_filename`
        workflow.transform(
            name='splitbyline',
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                1,
                mgd.TempOutputFile('input_filename', 'byline')))

        # Write the list of chunks (line indexes) on the axis `byline` produced
        # by splitting the input file
        workflow.transform(
            name='writelist',
            func='pypeliner.tests.tasks.write_list',
            args=(
                mgd.InputChunks('byline'),
                mgd.OutputFile(self.output_filename),))

        self.run_workflow(workflow, cleanup=False)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['01234567'])

    def test_multiple_object_split(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Read input file and store in managed input object, which is
        # a string of the file contents
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename),))

        # Split the string by line and store as a new object
        workflow.transform(
            name='splitbyline',
            func='pypeliner.tests.tasks.split_by_line',
            ret=mgd.TempOutputObj('input_data', 'byline'),
            args=(mgd.TempInputObj('input_data'),))

        # Split each of the resulting strings by character and
        # output as single character strings
        workflow.transform(
            name='splitbychar',
            axes=('byline',),
            func='pypeliner.tests.tasks.split_by_char',
            ret=mgd.TempOutputObj('input_data', 'byline', 'bychar'),
            args=(mgd.TempInputObj('input_data', 'byline'),))

        # Transform each single character string, appending a `-` character
        workflow.transform(
            name='do',
            axes=('byline', 'bychar'),
            func='pypeliner.tests.tasks.do_stuff',
            ret=mgd.TempOutputObj('output_data', 'byline', 'bychar'),
            args=(mgd.TempInputObj('input_data', 'byline', 'bychar').prop('some_string'),))

        # Merge modified strings along the `bychar` axis
        workflow.transform(
            name='mergebychar',
            axes=('byline',),
            func='pypeliner.tests.tasks.merge_stuff',
            ret=mgd.TempOutputObj('output_data', 'byline'),
            args=(mgd.TempInputObj('output_data', 'byline', 'bychar'),))

        # Merge modified strings along the `byline` axis
        workflow.transform(
            name='mergebyline',
            func='pypeliner.tests.tasks.merge_stuff',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('output_data', 'byline'),))

        # Write the merged string to an output file
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.write_stuff',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile(self.output_filename),))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['l-i-n-e-1-l-i-n-e-2-l-i-n-e-3-l-i-n-e-4-l-i-n-e-5-l-i-n-e-6-l-i-n-e-7-l-i-n-e-8-'])

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Redo the same steps and ensure that each step is skipped because each
        # step is up to date
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename),))
        workflow.transform(
            name='splitbyline',
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('input_data', 'byline'),
            args=(mgd.TempInputObj('input_data'),))
        workflow.transform(
            name='splitbychar',
            axes=('byline',),
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('input_data', 'byline', 'bychar'),
            args=(mgd.TempInputObj('input_data', 'byline'),))
        workflow.transform(
            name='do',
            axes=('byline', 'bychar'),
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('output_data', 'byline', 'bychar'),
            args=(mgd.TempInputObj('input_data', 'byline', 'bychar').prop('some_string'),))
        workflow.transform(
            name='mergebychar',
            axes=('byline',),
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('output_data', 'byline'),
            args=(mgd.TempInputObj('output_data', 'byline', 'bychar'),))
        workflow.transform(
            name='mergebyline',
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('output_data', 'byline'),))
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.do_assert',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile(self.output_filename),))

        self.run_workflow(workflow)

    def test_multiple_file_split(self):

        workflow = pypeliner.workflow.Workflow(default_ctx=self.ctx)

        # Split input file into 4 lines per output file (axis `byline_a`)
        workflow.transform(
            name='split_byline_a',
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                4,
                mgd.TempOutputFile('input_data', 'byline_a')))

        # Split again, this time with 2 lines per output file (axis `byline_b`)
        workflow.transform(
            name='split_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.TempInputFile('input_data', 'byline_a'),
                2,
                mgd.TempOutputFile('input_data', 'byline_a', 'byline_b')))

        # Modify each file independently, adding the instance of this job on the
        # `byline_a` axis
        workflow.transform(
            name='do',
            axes=('byline_a', 'byline_b'),
            func='pypeliner.tests.tasks.do_file_stuff',
            args=(
                mgd.TempInputFile('input_data', 'byline_a', 'byline_b'),
                mgd.TempOutputFile('output_data', 'byline_a', 'byline_b'),
                mgd.InputInstance('byline_a')))

        # Merge along the `byline_b` axis
        workflow.transform(
            name='merge_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('output_data', 'byline_a', 'byline_b'),
                mgd.TempOutputFile('output_data', 'byline_a')))

        # Merge along the `byline_a` axis
        workflow.transform(
            name='merge_byline_a',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('output_data', 'byline_a'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow, cleanup=False)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['00line1\n', '10line2\n', '00line3\n', '10line4\n', '01line5\n', '11line6\n', '01line7\n', '11line8\n'])

        tmp_input_data_filenames = [os.path.join(pipeline_dir, 'tmp/byline_a/0/input_data'), os.path.join(pipeline_dir, 'tmp/byline_a/1/input_data')]
        tmp_data_checks = [['line1\n', 'line2\n', 'line3\n', 'line4\n'], ['line5\n', 'line6\n', 'line7\n', 'line8\n']]
        for tmp_input_data_filename, tmp_data_check in zip(tmp_input_data_filenames, tmp_data_checks):
            with open(tmp_input_data_filename, 'r') as tmp_input_data_file:
                tmp_input_data = tmp_input_data_file.readlines()
                self.assertEqual(tmp_input_data, tmp_data_check)

    def test_rerun_simple(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        self.assertFalse(os.path.exists(self.output_filename))

        # Modify input file, append `!` to each line
        workflow.transform(
            name='step1',
            func='pypeliner.tests.tasks.append_to_lines',
            args=(
                mgd.InputFile(self.input_filename),
                '!',
                mgd.TempOutputFile('appended')))

        # Copy the file
        workflow.transform(
            name='step2',
            func='pypeliner.tests.tasks.copy_file',
            args=(
                mgd.TempInputFile('appended'),
                mgd.TempOutputFile('appended_copy')))

        # This job should copy the file again but does nothing, raising
        # an exception when the pipeline is run
        workflow.transform(
            name='step3',
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('appended_copy'),
                mgd.OutputFile(self.output_filename)))

        self.assertRaises(pypeliner.scheduler.PipelineException, self.run_workflow, workflow, cleanup=False)

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Redo the previous steps, ensuring the first two steps are not
        # run since their dependencies are up to date, and make sure the final
        # copy is done correctly
        workflow.transform(
            name='step1',
            func='pypeliner.tests.tasks.do_assert',
            args=(
                mgd.InputFile(self.input_filename),
                '!',
                mgd.TempOutputFile('appended')))
        workflow.transform(
            name='step2',
            func='pypeliner.tests.tasks.do_assert',
            args=(
                mgd.TempInputFile('appended'),
                mgd.TempOutputFile('appended_copy')))
        workflow.transform(
            name='step3',
            func='pypeliner.tests.tasks.copy_file',
            args=(
                mgd.TempInputFile('appended_copy'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow, cleanup=True)

        # The temporary files should have been cleaned up
        self.assertFalse(os.path.exists(os.path.join(pipeline_dir, 'tmp/appended')))
        self.assertFalse(os.path.exists(os.path.join(pipeline_dir, 'tmp/appended_copy')))

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1!\n', 'line2!\n', 'line3!\n', 'line4!\n', 'line5!\n', 'line6!\n', 'line7!\n', 'line8!\n'])

    def test_repopulate(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        self.assertFalse(os.path.exists(self.output_filename))

        # Modify input file, append `!` to each line
        workflow.transform(
            name='step1',
            func='pypeliner.tests.tasks.append_to_lines',
            args=(
                mgd.InputFile(self.input_filename),
                '!',
                mgd.TempOutputFile('appended')))

        # Copy the file
        workflow.transform(
            name='step2',
            func='pypeliner.tests.tasks.copy_file',
            args=(
                mgd.TempInputFile('appended'),
                mgd.TempOutputFile('appended_copy')))

        # This job should copy the file again but does nothing, raising
        # an exception when the pipeline is run
        workflow.transform(
            name='step3',
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('appended_copy'),
                mgd.OutputFile(self.output_filename)))

        self.assertRaises(pypeliner.scheduler.PipelineException, self.run_workflow, workflow)

        self.assertFalse(os.path.exists(os.path.join(pipeline_dir, 'tmp/appended')))

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Rerun the same pipeline in repopulate mode, this time the final
        # copy is done correctly
        workflow.transform(
            name='step1',
            func='pypeliner.tests.tasks.append_to_lines',
            args=(
                mgd.InputFile(self.input_filename),
                '!',
                mgd.TempOutputFile('appended')))
        workflow.transform(
            name='step2',
            func='pypeliner.tests.tasks.copy_file',
            args=(
                mgd.TempInputFile('appended'),
                mgd.TempOutputFile('appended_copy')))
        workflow.transform(
            name='step3',
            func='pypeliner.tests.tasks.copy_file',
            args=(
                mgd.TempInputFile('appended_copy'),
                mgd.OutputFile(self.output_filename)))

        repopulate_runskip = pypeliner.runskip.BasicRunSkip(repopulate=True)
        self.run_workflow(workflow, cleanup=False, runskip=repopulate_runskip)

        # The temporary files should have been cleaned up
        self.assertTrue(os.path.exists(os.path.join(pipeline_dir, 'tmp/appended')))
        self.assertTrue(os.path.exists(os.path.join(pipeline_dir, 'tmp/appended_copy')))

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1!\n', 'line2!\n', 'line3!\n', 'line4!\n', 'line5!\n', 'line6!\n', 'line7!\n', 'line8!\n'])

    def test_rerun_multiple_file_split(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Split input file into 4 lines per output file (axis `byline_a`)
        workflow.transform(
            name='split_byline_a',
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                4,
                mgd.TempOutputFile('input_data', 'byline_a')))

        # Split again, this time with 2 lines per output file (axis `byline_b`)
        workflow.transform(
            name='split_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.TempInputFile('input_data', 'byline_a'),
                2,
                mgd.TempOutputFile('input_data', 'byline_a', 'byline_b')))

        # Modify each file independently, adding the instance of this job on the
        # `byline_a` axis, fail here
        workflow.transform(
            name='do',
            axes=('byline_a', 'byline_b'),
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('input_data', 'byline_a', 'byline_b'),
                '!',
                mgd.TempOutputFile('output_data', 'byline_a', 'byline_b')))

        # Merge along the `byline_b` axis
        workflow.transform(
            name='merge_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('output_data', 'byline_a', 'byline_b'),
                mgd.TempOutputFile('output_data', 'byline_a')))

        # Merge along the `byline_a` axis
        workflow.transform(
            name='merge_byline_a',
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('output_data', 'byline_a'),
                mgd.OutputFile(self.output_filename)))

        self.assertRaises(pypeliner.scheduler.PipelineException, self.run_workflow, workflow)

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Split input file into 4 lines per output file (axis `byline_a`)
        workflow.transform(
            name='split_byline_a',
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                4,
                mgd.TempOutputFile('input_data', 'byline_a')))

        # Split again, this time with 2 lines per output file (axis `byline_b`)
        workflow.transform(
            name='split_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.TempInputFile('input_data', 'byline_a'),
                2,
                mgd.TempOutputFile('input_data', 'byline_a', 'byline_b')))

        # Modify each file independently, adding the instance of this job on the
        # `byline_a` axis
        workflow.transform(
            name='do',
            axes=('byline_a', 'byline_b'),
            func='pypeliner.tests.tasks.append_to_lines',
            args=(
                mgd.TempInputFile('input_data', 'byline_a', 'byline_b'),
                '!',
                mgd.TempOutputFile('output_data', 'byline_a', 'byline_b')))

        # Merge along the `byline_b` axis, fail here
        workflow.transform(
            name='merge_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('output_data', 'byline_a', 'byline_b'),
                mgd.TempOutputFile('output_data', 'byline_a')))

        # Merge along the `byline_a` axis
        workflow.transform(
            name='merge_byline_a',
            func='pypeliner.tests.tasks.do_nothing',
            args=(
                mgd.TempInputFile('output_data', 'byline_a'),
                mgd.OutputFile(self.output_filename)))

        self.assertRaises(pypeliner.scheduler.PipelineException, self.run_workflow, workflow)

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Split input file into 4 lines per output file (axis `byline_a`)
        workflow.transform(
            name='split_byline_a',
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.InputFile(self.input_filename),
                4,
                mgd.TempOutputFile('input_data', 'byline_a')))

        # Split again, this time with 2 lines per output file (axis `byline_b`)
        workflow.transform(
            name='split_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.split_file_byline',
            args=(
                mgd.TempInputFile('input_data', 'byline_a'),
                2,
                mgd.TempOutputFile('input_data', 'byline_a', 'byline_b')))

        # Modify each file independently, adding the instance of this job on the
        # `byline_a` axis
        workflow.transform(
            name='do',
            axes=('byline_a', 'byline_b'),
            func='pypeliner.tests.tasks.append_to_lines',
            args=(
                mgd.TempInputFile('input_data', 'byline_a', 'byline_b'),
                '!',
                mgd.TempOutputFile('output_data', 'byline_a', 'byline_b')))

        # Merge along the `byline_b` axis
        workflow.transform(
            name='merge_byline_b',
            axes=('byline_a',),
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('output_data', 'byline_a', 'byline_b'),
                mgd.TempOutputFile('output_data', 'byline_a')))

        # Merge along the `byline_a` axis
        workflow.transform(
            name='merge_byline_a',
            func='pypeliner.tests.tasks.merge_file_byline',
            args=(
                mgd.TempInputFile('output_data', 'byline_a'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1!\n', 'line2!\n', 'line3!\n', 'line4!\n', 'line5!\n', 'line6!\n', 'line7!\n', 'line8!\n'])

    def test_object_identical(self):

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        # Read data into a managed object
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename),))

        # Extract a property of the managed object, modify it
        # and store the result in another managed object
        workflow.transform(
            name='do',
            func='pypeliner.tests.tasks.do_stuff',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('input_data').prop('some_string'),))

        # Write the object to an output file
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.write_stuff',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile(self.output_filename)))

        self.run_workflow(workflow, cleanup=False)

        with open(self.output_filename, 'r') as output_file:
            output = output_file.readlines()

        self.assertEqual(output, ['line1\n', 'line2\n', 'line3\n', 'line4\n', 'line5\n', 'line6\n', 'line7\n', 'line8-'])

        print ('restarting')
        time.sleep(1)

        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        shutil.copyfile(self.input_filename, self.input_filename+'.tmp')

        # Read the same data into a managed object
        workflow.transform(
            name='read',
            func='pypeliner.tests.tasks.read_stuff',
            ret=mgd.TempOutputObj('input_data'),
            args=(mgd.InputFile(self.input_filename+'.tmp'),))

        # Extract a property of the managed object, modify it
        # and store the result in another managed object
        workflow.transform(
            name='do',
            func='pypeliner.tests.tasks.do_assert',
            ret=mgd.TempOutputObj('output_data'),
            args=(mgd.TempInputObj('input_data').prop('some_string'),))

        # Write the object to an output file
        workflow.transform(
            name='write',
            func='pypeliner.tests.tasks.do_assert',
            args=(
                mgd.TempInputObj('output_data'),
                mgd.OutputFile(self.output_filename),))

        self.run_workflow(workflow)

    def test_regenerate_temporaries_remixt_1(self):
        workflow = pypeliner.workflow.Workflow(ctx=self.ctx)

        workflow.setobj(obj=mgd.TempOutputObj('config'), value={})

        workflow.transform(
            name='calc_fragment_stats',
            func='pypeliner.tests.tasks.calculate_fragment_stats',
            ret=mgd.TempOutputObj('fragstats'),
            args=(mgd.InputFile(self.input_1_filename),)
        )

        workflow.transform(
            name='sample_gc',
            func='pypeliner.tests.tasks.sample_gc',
            args=(
                mgd.TempOutputFile('gcsamples.tsv'),
                mgd.InputFile(self.input_1_filename),
                mgd.TempInputObj('fragstats').prop('fragment_mean'),
                mgd.TempInputObj('config'),
                '',
            )
        )

        workflow.transform(
            name='gc_lowess',
            func='pypeliner.tests.tasks.gc_lowess',
            args=(
                mgd.TempInputFile('gcsamples.tsv'),
                mgd.TempOutputFile('gcloess.tsv'),
                mgd.TempOutputFile('gctable.tsv'),
            )
        )

        workflow.transform(
            name='split_segments',
            func='pypeliner.tests.tasks.split_table',
            args=(
                mgd.TempOutputFile('segments.tsv', 'segment_rows_idx'),
                mgd.InputFile(self.input_2_filename),
                100,
            ),
        )

        workflow.transform(
            name='gc_map_bias',
            axes=('segment_rows_idx',),
            func='pypeliner.tests.tasks.gc_map_bias',
            args=(
                mgd.TempInputFile('segments.tsv', 'segment_rows_idx'),
                mgd.TempInputObj('fragstats').prop('fragment_mean'),
                mgd.TempInputObj('fragstats').prop('fragment_stddev'),
                mgd.TempInputFile('gcloess.tsv'),
                mgd.TempOutputFile('biases.tsv', 'segment_rows_idx'),
                mgd.TempInputObj('config'),
                '',
            )
        )

        workflow.transform(
            name='merge_biases',
            func='pypeliner.tests.tasks.merge_tables',
            args=(
                mgd.TempOutputFile('biases.tsv'),
                mgd.TempInputFile('biases.tsv', 'segment_rows_idx'),
            ),
        )

        workflow.transform(
            name='biased_length',
            func='pypeliner.tests.tasks.biased_length',
            args=(
                mgd.OutputFile(self.output_filename),
                mgd.TempInputFile('biases.tsv'),
            ),
        )

        self.run_workflow(workflow)

        time.sleep(1)
        os.utime(self.input_2_filename, None)

        print ('restarting')

        self.run_workflow(workflow)


if __name__ == '__main__':
    unittest.main()



import os
import errno

import pypeliner.execqueue.base


class SubProcessJobQueue(pypeliner.execqueue.base.JobQueue):
    """ Abstract class for a queue of jobs run using subprocesses.  Maintains
    a list of running jobs, with the ability to wait for jobs and return
    completed jobs.  Requires override of the create method.
    """
    def __init__(self, modules=None, **kwargs):
        self.modules = modules
        self.jobs = dict()
        self.pid_names = dict()
        self.pid_returncodes = dict()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        pass

    def create(self, ctx, name, sent, temps_dir):
        raise NotImplementedError()

    def send(self, ctx, name, sent, temps_dir):
        submitted = self.create(ctx, name, sent, temps_dir)
        self.jobs[name] = submitted
        self.pid_names[submitted.process.pid] = name

    def wait(self, immediate=False):
        while True:
            if immediate:
                process_id, returncode = os.waitpid(-1, os.WNOHANG)
                if process_id is None:
                    return None
            else:
                try:
                    process_id, returncode = os.wait()
                except OSError as e:
                    if e.errno == errno.EINTR:
                        continue
                    else:
                        raise
            try:
                name = self.pid_names.pop(process_id)
            except KeyError:
                continue
            self.pid_returncodes[name] = returncode
            return name

    def receive(self, name):
        job = self.jobs.pop(name)
        job.finalize(self.pid_returncodes.pop(name))
        return job.received

    @property
    def length(self):
        return len(self.jobs)

    @property
    def empty(self):
        return self.length == 0




import boto3
import botocore

class AwsSimpleStorageService(object):
    """
    :param access_id: AWS IAM access id (optional)
    :type access_id: str
    :param access_key: AWS IAM access key (optional)
    :type access_key: str
    :param location: AWS datacenter location (optional)
    :type location: str
    :param multipart_threshold: The transfer size threshold for which multipart
    uploads, downloads, and copies will automatically be triggered.
    :type multipart_threshold: int
    :param num_download_attempts: number of retries
    :type num_download_attempts: int
    :param max_concurrency: number of concurrent downloads(threads)
    :type max_concurrency: int
    :param multipart_chunksize: The partition size of each part for a multipart transfer.
    :type multipart_chunksize: int
    :param max_io_queue: The maximum amount of read parts that can be queued in memory to be written for a download
    :type max_io_queue: int
    :param io_chunksize: The max size of each chunk in the io queue
    :type io_chunksize: int
    """

    def __init__(
            self,
            access_id=None,
            access_key=None,
            location=None,
            multipart_threshold=512*1024*1024,
            num_download_attempts=5,
            max_concurrency=1,
            multipart_chunksize=512*1024*1024,
            max_io_queue=100,
            io_chunksize=1024*1024
    ):
        self.client = boto3.client('s3', aws_access_key_id=access_id, aws_secret_access_key=access_key)
        self.resource = boto3.resource('s3', aws_access_key_id=access_id, aws_secret_access_key=access_key)
        self.location = location

        transfer_config = boto3.s3.transfer.TransferConfig(
            multipart_threshold=multipart_threshold,
            num_download_attempts=num_download_attempts,
            max_concurrency=max_concurrency,
            multipart_chunksize=multipart_chunksize,
            max_io_queue=max_io_queue,
            io_chunksize=io_chunksize
        )
        self.transfer = boto3.s3.transfer.S3Transfer(
            client=self.client,
            config=transfer_config
        )

    def create_bucket(self, bucketname):
        """
        create a new S3 bucket
        :param bucketname: name of bucket
        :type bucketname: str
        """
        if self.bucket_exists(bucketname):
            return
        response = self.client.create_bucket(
            ACL='private',
            Bucket=bucketname,
            CreateBucketConfiguration={
                'LocationConstraint': self.location
            },
            ObjectLockEnabledForBucket=True,
        )

    def create_signed_url(self, bucket, key):
        url = self.client.generate_presigned_url(
            ClientMethod='get_object',
            Params={
                'Bucket': bucket,
                'Key': key
            },
            ExpiresIn=604800
        )

        return url


    def bucket_exists(self, bucket):
        """
        check if a bucket exists
        :param bucket: bucket name
        :type bucket: str
        :return: True if exists
        :rtype: bool
        """
        bucket = self.resource.Bucket(bucket)
        if bucket.creation_date:
            return True
        else:
            return False

    def key_exists(self, bucket, key):
        """
        check if an object exists
        :param bucket: bucket name
        :type bucket: str
        :param key: object name
        :type key: str
        :return: True if exists
        :rtype: bool
        """
        bucket = self.resource.Bucket(bucket)
        s3object = bucket.Object(key)
        try:
            s3object.load()
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                # The object does not exist.
                return False
            else:
                # Something else has gone wrong.
                raise
        else:
            return True

    def upload_from_file(self, source, destination, bucket):
        """
        upload file to S3. uses multipart for big files
        :param source: filename on filesystem
        :type source: str
        :param destination: key in S3
        :type destination: str
        :param bucket: bucket in S3
        :type bucket: str
        """
        self.transfer.upload_file(source, bucket, destination)

    def get_file_size(self, bucket, filename):
        """
        get size of object in S3
        :param bucket: bucket name
        :type bucket: str
        :param filename: object name
        :type filename: str
        :return: size
        :rtype: int
        """
        if not self.bucket_exists(bucket):
            raise Exception()
        if not self.key_exists(bucket, filename):
            raise Exception()

        bucket = self.resource.Bucket(bucket)
        s3object = bucket.Object(filename)
        return s3object.content_length

    def delete_key(self, bucket, key):
        """
        delete an object
        :param bucket: bucket name
        :type bucket: str
        :param key: object name
        :type key: str
        """
        self.resource.Object(bucket, key).delete()

    def download_to_path(self, source, bucket, destination):
        """
        download from S3. uses multipart for big files
        :param source: object name in S3
        :type source: str
        :param bucket: bucket name in S3
        :type bucket: str
        :param destination: file path in local filesystem
        :type destination: str
        """

        try:
            self.transfer.download_file(bucket, source, destination)
        except botocore.exceptions.ClientError as e:
            if e.response['Error']['Code'] == "404":
                print (source,bucket, destination)
                raise Exception("The object does not exist.")
            else:
                raise

    def set_metadata(self, bucket, key, metadata_key, metadata_value):
        """
        set metadata key value pair for an object in S3
        :param bucket: bucket name
        :type bucket: str
        :param key: object name
        :type key: str
        :param metadata_key: key to store the metadata under
        :type metadata_key:  str
        :param metadata_value: value for the key
        :type metadata_value: str
        """
        copy_source = {
            'Bucket': bucket,
            'Key': key
        }
        self.client.copy(
            copy_source, bucket, key,
            ExtraArgs={
                "Metadata": {
                    metadata_key: metadata_value
                },
                "MetadataDirective": "REPLACE"
            }
        )

    def get_metadata(self, bucket, key, metadata_key):
        """
        get value for provided key in object metadata in S3
        :param bucket: bucket name
        :type bucket: str
        :param key: object name
        :type key: str
        :param metadata_key: key
        :type metadata_key: str
        :return: value for key
        :rtype:  str
        """
        key_metadata = self.client.head_object(Bucket=bucket, Key=key)
        custom_metadata = key_metadata.get('Metadata')
        if not custom_metadata:
            return None
        createtime = custom_metadata.get(metadata_key)
        return createtime

    def get_last_modified(self, bucket, key):
        """
        get time the object was last modified
        :param bucket: bucket name
        :type bucket: str
        :param key: object name
        :type key: str
        :return: time of last change(format: %Y/%m/%d-%H:%M:%S)
        :rtype: str
        """
        timeobj = self.resource.Bucket(bucket).Object(key).last_modified
        return timeobj.strftime('%Y/%m/%d-%H:%M:%S')



import functools
import hashlib
import logging
import os
import shutil

import pypeliner.commandline as cli
import yaml


class CondaSandbox(object):

    def __init__(self, channels=None, packages=None, pip_packages=None):
        if channels is None:
            channels = []
        if packages is None:
            packages = []
        if pip_packages is None:
            pip_packages = []
        self.channels = list(channels)
        self.packages = list(packages)
        self.pip_packages = list(pip_packages)
        self.prefix = None
        self.logger = logging.getLogger('pypeliner.sandbox')

    @property
    def config(self):
        return {'channels': self.channels, 'packages': self.packages, 'pip_packages': self.pip_packages}

    def _get_prefix(self):
        m = hashlib.md5()
        m.update(yaml.dump(self.config).encode())
        return m.hexdigest()

    def create_conda_env(self, env_dir):
        self.prefix = os.path.join(env_dir, self._get_prefix())
        env_config_file = os.path.join(self.prefix, 'sandbox_config.yaml')
        # TODO: Probably want some error checking on packages etc.
        if os.path.exists(self.prefix):
            with open(env_config_file, 'r') as fh:
                env_config = yaml.safe_load(fh)
            if (env_config['channels'] == self.channels) and (set(env_config['packages']) == set(self.packages)):
                return False
            else:
                shutil.rmtree(self.prefix)
        self._write_log()
        self._run_conda()
        with open(env_config_file, 'w') as out_fh:
            yaml.dump(self.config, out_fh)
        return True

    def wrap_function(self, func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            if self.prefix is None:
                raise Exception('Sandbox not initialized.')
            self._set_env()
            return func(*args, **kwargs)

        return wrapper

    def _run_conda(self):
        # TODO: Should add some retry logic in case of network timeouts
        cmd = ['conda', 'create', '--no-channel-priority', '--yes', '--quiet', '--prefix', self.prefix]
        for channel in self.channels:
            cmd.extend(['-c', channel])
        cmd.extend(self.packages + ['pip', ])
        cmd.extend(['>', '/dev/null'])
        cli.execute(*cmd)
        self._set_env()
        pip_exe = os.path.join(self.prefix, 'bin', 'pip')
        for p in self.pip_packages:
            cli.execute(pip_exe, 'install', p, '>', '/dev/null')

    def _set_env(self):
        os.environ['PATH'] = os.pathsep.join([os.path.join(self.prefix, 'bin'), os.environ['PATH']])
        os.environ['CONDA_PREFIX'] = self.prefix

    def _write_log(self):
        self.logger.info('sandbox {0} creating in {1}'.format(
            os.path.basename(self.prefix), self.prefix))
        self.logger.info('sandbox {0} conda channels {1}'.format(
            os.path.basename(self.prefix), ','.join(sorted(self.channels))))
        self.logger.info('sandbox {0} conda packages {1}'.format(
            os.path.basename(self.prefix), ','.join(sorted(self.packages))))
        self.logger.info('sandbox {0} pip packages {1}'.format(
            os.path.basename(self.prefix), ','.join(sorted(self.pip_packages))))
